import argparse
from gooey import Gooey

@Gooey()
def process_options():
  parser = argparse.ArgumentParser()
  parser.add_argument("-v", action="count", help='verbose mode', default=0)
  return parser.parse_args()


if __name__ == '__main__':
  args = process_options()
  print(args.v)


import argparse
from gooey import Gooey

@Gooey(poll_external_updates=True, dump_build_config=True)
def main():

    parser = argparse.ArgumentParser(description='Process some integers.')
    parser.add_argument('integers', metavar='N', type=int, nargs='+',
                       help='an integer for the accumulator',
                        default='')
    parser.add_argument('-foo', type=int, nargs='+',
                        help='fooobie',
                        default='')
    parser.add_argument('--sum', dest='accumulate', action='store_const',
                       const=sum, default=max,
                       help='sum the integers (default: find the max)')

    args = parser.parse_args()
    print( args.accumulate(args.integers))

if __name__=="__main__":
    import sys
    import json
    if 'gooey-seed-ui' in sys.argv:
        print(json.dumps({'integers': '666'}))
    else:
        main()



from argparse import ArgumentParser
from gooey import Gooey
import time
import signal


running = [True]

@Gooey(shutdown_signal=signal.CTRL_BREAK_EVENT,
       program_name='Graceful Stopification',
       disable_progress_bar_animation=True)
def main_argparse():
    parser = ArgumentParser(description='Sum two integers.')
    # parser.add_argument('integers', type=int, nargs=2)
    # signal.signal(signal.SIGTERM, logit)
    # signal.signal(signal.SIGINT , logit)
    # signal.signal(signal.SIGBREAK, logit)
    # signal.signal(signal.SIGILL, logit)

    try:
        args = parser.parse_args()
        end = time.time() + 101111
        while time.time() < end and running[0]:
            print("Just try and stop me!")
            # print('waiting...', time.time(), running)
            time.sleep(0.2)

        if not running:
            print("stopped due to Ctrl+C")
    except (KeyboardInterrupt, SystemExit):
        print("HOLY SHIT!!!")


def logit(*args):
    print(args, 'logit')
    # raise Exception("get fucked!!!!")
    running[0] = False
    # print("HELLLLLLL")
    print("Shutting down!")

if __name__ == "__main__":
    # main_argparse()
    main_argparse()


import wx  # type: ignore

from gooey.gui import formatters
from gooey.gui.components.widgets.bases import TextContainer
from gooey.python_bindings import types as t


class Slider(TextContainer):
    """
    An integer input field
    """
    widget_class = wx.Slider
    def getWidget(self, *args, **options):
        widget = self.widget_class(self,
                                   minValue=self._options.get('min', 0),
                                   maxValue=self._options.get('max', 100),
                                   style=wx.SL_MIN_MAX_LABELS | wx.SL_VALUE_LABEL)
        return widget

    def getWidgetValue(self):
        return self.widget.GetValue()

    def setValue(self, value):
        self.widget.SetValue(value)

    def formatOutput(self, metatdata, value):
        return formatters.general(metatdata, str(value))

    def getUiState(self) -> t.FormField:
        widget: wx.Slider = self.widget
        return t.Slider(
            id=self._id,
            type=self.widgetInfo['type'],
            value=self.getWidgetValue(),
            min=widget.GetMin(),
            max=widget.GetMax(),
            error=self.error.GetLabel() or None,
            enabled=self.IsEnabled(),
            visible=self.IsShown()
        )



from random import random
from typing import List

from gooey import Gooey, GooeyParser, constants, Events
import os

from gooey.python_bindings.types import PublicGooeyState
from gooey.python_bindings.types import FormField


save_dir = os.path.join(os.path.dirname(__file__), 'saves')

def grab_latest(args, state: PublicGooeyState):
    form: List[FormField] = state['active_form']
    for item in form:
        if item['id'] == 'length':
            item['value'] = 'Good job, mang!'
        if item['id'] == 'options':
            item['choices'] = os.listdir(save_dir)
    state['active_form'] = form
    return state

def handle_error(args, state: PublicGooeyState):
    form: List[FormField] = state['active_form']
    for item in form:
        if item['id'] == 'length':
            item['value'] = 'no way, bro'
    state['active_form'] = form
    return state


def nope(x):
    if int(x) == 10:
        return int(x)
    else:
        raise Exception('omg. Get fucked!')


@Gooey(optional_cols=2,
        program_name="Elapsed / Remaining Timer on Progress in Gooey",
        progress_regex=r"^progress: (?P<current>\d+)/(?P<total>\d+)$",
        progress_expr="current / total * 100",
        use_events=[Events.VALIDATE_FORM, Events.ON_SUCCESS, Events.ON_ERROR],
        show_sidebar=True,
        show_preview_warning=False,
        timing_options={
            'show_time_remaining':True,
            'hide_time_remaining_on_complete':False
        },
        menu=[{'name': 'Help', 'items': [{'type': 'Link', 'menuTitle': 'Visit Our Site',
                                         'url': 'https://github.com/chriskiehl/Gooey'}]}]
        )
def parse_args():
    print('hello!')
    print('starting run!')
    prog_descrip = 'Elapsed / Remaining Timer on Progress in Gooey'
    parser = GooeyParser(
        description=prog_descrip,
        on_success=grab_latest,
        on_error=handle_error
    )

    from random import choices

    sub_parsers = parser.add_subparsers(help='commands', dest='command')

    range_parser = sub_parsers.add_parser('range')

    range_parser.add_argument('--length',default=10, type=nope, widget='MultiFileChooser')

    group = range_parser.add_argument_group('foobar')
    group.add_argument(
        '--options',
        choices=os.listdir(save_dir),
        nargs='*',
        widget='Listbox')

    # subgroup = group.add_argument_group("inner")
    # subgroup.add_argument('--schlength', default='55th street')
    #
    # strange_parser = sub_parsers.add_parser('strange')
    # strange_parser.add_argument('--length', default=10)
    #
    # strange_parser = sub_parsers.add_parser('mange')

    x = parser.parse_args()
    import time
    # raise Exception("FUCK!")
    with open(os.path.join(save_dir, f'Save-{int(time.time())}.txt'), 'w') as f:
        f.write('hello')
    print(x)
    return x


def compute_range(length):
    for i in range(length):
        # print("sometin")
        import time
        from random import randint
        time.sleep(0.5)
        print(f"progress: {i + 1}/{length}")

if __name__ == '__main__':
    conf = parse_args()
    if conf.command == 'range':
        compute_range(int(conf.length))


import os

import itertools

from gooey.gui.util.quoting import quote
from gooey.python_bindings.types import EnrichedItem, FormField
from gooey.gui.constants import VALUE_PLACEHOLDER, RADIO_PLACEHOLDER
from gooey.util.functional import assoc, associnMany


def value(field: FormField):
    if field['type'] in ['Checkbox', 'BlockCheckbox']:
        return field['checked']   # type: ignore
    elif field['type'] in ['Dropdown', 'Listbox', 'Counter']:
        return field['selected']   # type: ignore
    elif field['type'] == 'RadioGroup':
        if field['selected'] is not None:  # type: ignore
            return value(field['options'][field['selected']])  # type: ignore
        else:
            return None
    else:
        return field['value'] # type: ignore


def add_placeholder(field: FormField, placeholder=VALUE_PLACEHOLDER):
    """
    TODO: Docs about placeholders
    """
    if field['type'] in ['Checkbox', 'CheckBox', 'BlockCheckbox']:
        # there's no sane placeholder we can make for this one, as
        # it's kind of a nonsensical case: a required optional flag.
        # We set it to True here, which is equally nonsensical, but
        # ultimately will allow the validation to pass. We have no
        # way of passing a placeholder without even MORE monket patching
        # of the user's parser to rewrite the action type
        return assoc(field, 'checked', True)
    elif field['type'] in ['Dropdown', 'Listbox', 'Counter']:
        return assoc(field, 'selected', placeholder)
    elif field['type'] == 'RadioGroup':
        # We arbitrarily attach a placeholder for first RadioGroup option
        # and mark it as the selected one.
        return {
            **field,
            'selected': 0,
            'options': [
                add_placeholder(field['options'][0], placeholder=RADIO_PLACEHOLDER),  # type: ignore
                *field['options'][1:]  # type: ignore
            ]
        }
    else:
        return assoc(field, 'value', placeholder)


def formatArgument(item: EnrichedItem):
    if item['type'] in ['Checkbox', 'CheckBox', 'BlockCheckbox']:
        return checkbox(item['data'], value(item['field']))
    elif item['type'] == 'MultiFileChooser':
        return multiFileChooser(item['data'], value(item['field']))
    elif item['type'] == 'Textarea':
        return textArea(item['data'], value(item['field']))
    elif item['type'] == 'CommandField':
        return textArea(item['data'], value(item['field']))
    elif item['type'] == 'Counter':
        return counter(item['data'], value(item['field']))
    elif item['type'] == 'Dropdown':
        return dropdown(item['data'], value(item['field']))
    elif item['type'] == 'Listbox':
        return listbox(item['data'], value(item['field']))
    elif item['type'] == 'RadioGroup':
        selected = item['field']['selected']  # type: ignore
        if selected is not None:
            formField = item['field']['options'][selected]  # type: ignore
            argparseDefinition = item['data']['widgets'][selected]  # type: ignore
            return formatArgument(assoc(argparseDefinition, 'field', formField))  # type: ignore
        else:
            return None
    else:
        return general(item['data'], value(item['field']))


def placeholder(item: EnrichedItem):
    pass


def checkbox(metadata, value):
    return metadata['commands'][0] if value else None


def multiFileChooser(metadata, value):
    paths = ' '.join(quote(x) for x in value.split(os.pathsep) if x)
    if metadata['commands'] and paths:
        return u'{} {}'.format(metadata['commands'][0], paths)
    return paths or None


def textArea(metadata, value):
    if metadata['commands'] and value:
        return '{} {}'.format(metadata['commands'][0], quote(value.encode('unicode_escape')))
    else:
        return quote(value.encode('unicode_escape')) if value else ''


def commandField(metadata, value):
    if metadata['commands'] and value:
        return u'{} {}'.format(metadata['commands'][0], value)
    else:
        return value or None


def counter(metatdata, value):
    '''
    Returns
      str(option_string * DropDown Value)
      e.g.
      -vvvvv
    '''
    if not str(value).isdigit():
        return None
    command = str(metatdata['commands'][0]).strip()
    return ' '.join(itertools.repeat(command, int(value)))


def dropdown(metadata, value):
    if value == 'Select Option':
        return None
    elif metadata['commands'] and value:
        return u'{} {}'.format(metadata['commands'][0], quote(value))
    else:
        return quote(value) if value else ''


def listbox(meta, value):
    if meta['commands'] and value:
        return u'{} {}'.format(meta['commands'][0], ' '.join(map(quote, value)))
    else:
        return ' '.join(map(quote, value)) if value else ''


def general(metadata, value):
    if metadata.get('commands') and value:
        if not metadata.get('nargs'):
            v = quote(value)
        else:
            v = value
        return u'{0} {1}'.format(metadata['commands'][0], v)
    else:
        if not value:
            return None
        elif not metadata.get('nargs'):
            return quote(value)
        else:
            return value




from gooey import GooeyParser, Gooey
from pprint import pprint


@Gooey(dump_build_config=True)
def main():
    parser = GooeyParser(description="FFQ profile generation and auditing")

    g2 = parser.add_argument_group("g2", "Group 2")
    g1 = parser.add_argument_group("g1", "Group 1")

    g1.add_argument("abc", default="abc")
    g2.add_argument("def", default="def")
    parser.add_argument("hij", default="hij")

    args = parser.parse_args()

    pprint(args)


if __name__ == "__main__":
    main()



import unittest
from gooey.gui import cli


class TestCliStringCreation(unittest.TestCase):

    # TODO: exercise the formValidationCase (which will require tedious test data creation)
    def test_cli(self):
        print(cli.buildCliString('target', 'cmd', ['pos1', 'pos2'], ['-a 1', '-b 2']))

        positionals = [
            {'clitype': 'positional', 'cmd': 'pos1', 'required': True},
            {'clitype': 'positional', 'cmd': 'pos2', 'required': True}
        ]

        optionals = [
            {'clitype': 'optional', 'cmd': '-a 1', 'required': False},
            {'clitype': 'optional', 'cmd': '-b 2', 'required': False},
        ]

        # print(cli.formValidationCmd('target', 'cmd', positionals, optionals))


"""
All things Dynamic Updates & Validation.

Hear me all ye who enter!
=========================

This is a module of disgusting hacks and monkey patching. Control flow
is all over the place and a comprised of hodgepodge of various strategies.
This is all because Argparse's internal parsing design (a) really,
really, REALLY wants to fail and sys.exit at the first error it
finds, and (b) does these program ending validations at seemingly random
points throughout its code base. Meaning, there is no single centralized
validation module, class, or function which could be overridden in order to
achieve the desired behavior.

All that means is that it takes a fair amount of indirect, non-standard, and
gross monkey-patching to get Argparse to collect all its errors as it parses
rather than violently explode each time it finds one.

For additional background, see the original design here:
https://github.com/chriskiehl/Gooey/issues/755


"""
from argparse import ArgumentParser, _SubParsersAction, _MutuallyExclusiveGroup
from functools import wraps
from typing import Union, Any, Mapping, Dict, Callable

from gooey.python_bindings.types import Success, Failure, Try, InvalidChoiceException
from gooey.python_bindings.argparse_to_json import is_subparser
from gooey.util.functional import lift, identity, merge
from gooey.gui.constants import VALUE_PLACEHOLDER
from gooey.python_bindings.constants import Events
from gooey.python_bindings.coms import decode_payload
from gooey.gui.constants import RADIO_PLACEHOLDER

unexpected_exit_explanations = f'''
+=======================+
|Gooey Unexpected Error!|
+=======================+

Gooey encountered an unexpected error while trying to communicate 
with your program to process one of the {Events._fields} events.

These features are new and experimental! You may have encountered a bug! 

You can open a ticket with a small reproducible example here
https://github.com/chriskiehl/Gooey/issues
'''  # type: ignore


deserialize_failure_explanations = f'''
+==================================+
|Gooey Event Deserialization Error!|
+==================================+

Gooey was unable to deserialize the payload returned from your 
program when processing one of the {Events._fields} events. 

The payload *MUST* be in the `GooeyPublicState` schema. You can 
view the type information in `gooey.python_bindings.types.py`

Note, these features are new an experimental. This may be a bug on 
Gooey's side! 

You can open a ticket with a small reproducible example here:  
https://github.com/chriskiehl/Gooey/issues
'''


def check_value(registry: Dict[str, Exception], original_fn):
    """
    A Monkey Patch for `Argparse._check_value` which changes its
    behavior from one which throws an exception, to one which swallows
    the exception and silently records the failure.

    For certain argument types, Argparse calls a
    one-off `check_value` method. This method is inconvenient for us
    as it either returns nothing or throws an ArgumentException (thus leading
    to a sys.exit). Because our goal is to collect ALL
    errors for the entire parser, we must patch around this behavior.
    """
    @wraps(original_fn)
    def inner(self, action, value: Union[Any, Success, Failure]):
        def update_reg(_self, _action, _value):
            try:
                original_fn(_action, _value)
            except Exception as e:
                # check_value exclusively handles validating that the
                # supplied argument is a member of the `choices` set.
                # by default, it pops an exception containing all of the
                # available choices. However, since we're in a UI environment
                # all of that is redundant information. It's also *way too much*
                # information for things like FilterableDropdown. Thus we just
                # remap it to a 'simple' exception here.
                error = InvalidChoiceException("Selected option is not a valid choice")
                # IMPORTANT! note that this mutates the
                # reference that is passed in!
                registry[action.dest] = error

        # Inside of Argparse, `type_func` gets applied before the calls
        # to `check_value`. A such, depending on the type, this may already
        # be a lifted value.
        if isinstance(value, Success) and not isinstance(value, Failure):
            update_reg(self, action, value.value)
        elif isinstance(value, list) and all(x.isSuccess() for x in value):
            update_reg(self, action, [x.value for x in value])
        else:
            update_reg(self, action, value)
    return inner


def patch_args(*args, **kwargs):
    def inner(parser):
        return patch_argument(parser, *args, **kwargs)
    return inner

def patch_argument(parser, *args, **kwargs):
    """
    Mutates the supplied parser to append the arguments (args, kwargs) to
    the root parser and all subparsers.

    Example: `patch_argument(parser, '--ignore-gooey', action='store_true')

    This is used to punch additional cli arguments into the user's
    existing parser definition. By adding our arguments everywhere it allows
    us to use the `parse_args` machinery 'for free' without needing to
    worry about context shifts (like a repeated `dest` between subparsers).
    """
    parser.add_argument(*args, **kwargs)
    subparsers = list(filter(is_subparser, parser._actions))
    if subparsers:
        for sub in subparsers[0].choices.values():  # type: ignore
            patch_argument(sub, *args, **kwargs)
    return parser


def patch_all_parsers(patch_fn: Callable[[ArgumentParser], None], parser):
    subparsers = list(filter(is_subparser, parser._actions))
    if subparsers:
        for sub in subparsers[0].choices.values():  # type: ignore
            patch_all_parsers(patch_fn, sub)
    return parser


def recursively_patch_parser(parser, fn, *args):
    fn(parser, *args)
    subparsers = list(filter(is_subparser, parser._actions))
    if subparsers:
        for sub in subparsers[0].choices.values():  # type: ignore
            recursively_patch_parser(sub, fn, *args)
    return parser


def recursively_patch_actions(parser, fn):
    for action in parser._actions:
        if issubclass(type(action), _SubParsersAction):
            for subparser in action.choices.values():
                recursively_patch_actions(subparser, fn)
        else:
            fn(action)

def lift_action_type(action):
    """"""
    action.type = lift(action.type or identity)

def lift_actions_mutating(parser):
    """
    Mutates the supplied parser to lift all of its (likely) partial
    functions into total functions. See module docs for additional
    background. TL;DR: we have to "trick" Argparse into thinking
    every value is valid so that it doesn't short-circuit and sys.exit
    when it encounters a validation error. As such, we wrap everything
    in an Either/Try, and defer deciding the actual success/failure of
    the type transform until later in the execution when we have control.
    """
    recursively_patch_actions(parser, lift_action_type)
    # for action in parser._actions:
    #     if issubclass(type(action), _SubParsersAction):
    #         for subparser in action.choices.values():
    #             lift_actions_mutating(subparser)
    #     else:
    #         action.type = lift(action.type or identity)


def collect_errors(parser, error_registry: Dict[str, Exception], args: Dict[str, Try]) -> Dict[str, str]:
    """
    Merges all the errors from the Args mapping and error registry
    into a final dict.
    """
    # As is a theme throughout this module, to avoid Argparse
    # short-circuiting during parse-time, we pass a placeholder string
    # for required positional arguments which haven't yet been provided
    # by the user. So, what's happening here is that we're now collecting
    # all the args which have the placeholders so that we can flag them
    # all as required and missing.
    # Again, to be hyper clear, this is about being able to collect ALL
    # errors, versus just ONE error (Argparse default).
    required_but_missing = {k: 'This field is required'
                            for k, v in args.items()
                            if isinstance(v, Success) and v.value == VALUE_PLACEHOLDER}

    mutexes_required_but_missing = collect_mutex_errors(parser, args)

    errors = {k: str(v.error)
              for k, v in args.items()
              if v is not None and isinstance(v, Failure)}
    # Secondary errors are those which get frustratingly applied by
    # Argparse in a way which can't be easily tracked with patching
    # or higher order functions. See: `check_value` for more details.
    secondary = {k: str(e) for k, e in error_registry.items() if e}
    return merge(required_but_missing, errors, secondary, mutexes_required_but_missing)


def collect_mutex_errors(parser, args: Dict[str, Try]):
    """
    RadioGroups / MutuallyExclusiveGroup require extra care.
    Mutexes are not normal actions. They're not argument targets
    themselves, they have no `dest`, they're just parse-time containers
    for arguments. As such, there's no top-level argument destination
    we can tie a single error to. So, the strategy here is to mark _all_ of
    a radio group's children with an error if *any* of them are missing.

    It's a bit clunky, but what we've got to work with.
    """
    def dest_targets(group: _MutuallyExclusiveGroup):
        return [action.dest for action in group._group_actions]

    mutexes_missing = {dest for dest, v in args.items()
                     if isinstance(v, Success) and v.value == RADIO_PLACEHOLDER}

    return {dest: 'One of these must be provided'
            for group in parser._mutually_exclusive_groups
            for dest in dest_targets(group)
            # if the group is required and we've got one of its
            # children marked as missing
            if group.required and set(dest_targets(group)).intersection(mutexes_missing)}









def patch(obj, old_fn, new_fn):
    setattr(obj, old_fn, new_fn.__get__(obj, ArgumentParser))

def monkey_patch_check_value(parser, new_fn):
    parser._check_value = new_fn.__get__(parser, ArgumentParser)
    return parser

def monkey_patch(patcher, error_registry: Dict[str, Exception], parser):
    lift_actions_mutating(parser)
    patcher(parser)
    new_check_value = check_value(error_registry, parser._check_value)
    # https://stackoverflow.com/questions/28127874/monkey-patching-python-an-instance-method
    # parser._check_value = new_check_value.__get__(parser, ArgumentParser)

    return parser



def monkey_patch_for_form_validation(error_registry: Dict[str, Exception], parser):
    """
    Applies all the crufty monkey patching required to
    process a validate_form event
    """
    lift_actions_mutating(parser)
    patch_argument(parser, '--gooey-validate-form', action='store_true')
    patch_argument(parser, '--gooey-state', action='store', type=decode_payload)
    new_check_value = check_value(error_registry, parser._check_value)
    recursively_patch_parser(parser, monkey_patch_check_value, new_check_value)
    # https://stackoverflow.com/questions/28127874/monkey-patching-python-an-instance-method
    # patch(parser, '_check_value', new_check_value)
    # parser._check_value = new_check_value.__get__(parser, ArgumentParser)
    return monkey_patch_check_value(parser, new_check_value)




from argparse import _StoreAction

from gooey import Gooey, GooeyParser
from python_bindings.control import Gooey1


def foo(*args):
    return [1, 2, 3]


from gooey import GooeyParser, Gooey


@Gooey(dump_build_config=True)
def main():
    parser = GooeyParser()
    group = parser.add_argument_group('Super cool group',
                                      gooey_options={'title': 'My gruop',
                                                     'metavar': 'Hello World!',
                                                     'columns': 2,
                                                     'full_width': 0,
                                                     'show_border': True})

    group.add_argument('--here-for-ui', gooey_options={'visible': False})
    group1 = group.add_argument_group('group1',
                                      gooey_options={'title': 'My gruop',
                                                     'metavar': 'Hello World!',
                                                     'columns': 2,
                                                     'full_width': 0,
                                                     'show_border': True})
    group1.add_argument('Rule_Description', help='desc')
    group1.add_argument('info', help='info')

    parser.parse_args()


if __name__ == '__main__':
    main()



import unittest
from argparse import ArgumentParser
from typing import Dict
from unittest.mock import MagicMock

from python_bindings.dynamics import patch_argument, monkey_patch_for_form_validation


class TestDynamicUpdates(unittest.TestCase):

    def tearDown(self):
        """
        Undoes the monkey patching after every tests
        """
        if hasattr(ArgumentParser, 'original_parse_args'):
            ArgumentParser.parse_args = ArgumentParser.original_parse_args

    def test_patch_argument(self):
        """
        Asserting that regardless of parser complexity, we attach our
        new argument at every level.
        """
        parser = ArgumentParser()
        subparsers = parser.add_subparsers()
        # multiple subparsers
        a = subparsers.add_parser('a')
        b = subparsers.add_parser('b')

        a.add_argument('--level-1')
        b.add_argument('--level-1')

        # deeply nested subparsers
        a_subparsers = a.add_subparsers()
        b_subparsers = b.add_subparsers()

        # nested args:
        a__nested = a_subparsers.add_parser('a1')
        b__nested = b_subparsers.add_parser('b1')

        a__nested.add_argument('--level-2')
        b__nested.add_argument('--level-2')

        # sanity check / showing the parser behavior
        # we've got two levels of parser nesting, each level
        # has some options available.
        mock = MagicMock()
        ArgumentParser.error = mock
        assert parser.parse_args('a --level-1 some-value'.split())
        assert parser.parse_args('b --level-1 some-value'.split())
        assert parser.parse_args('a a1 --level-2 some-value'.split())
        assert parser.parse_args('b b1 --level-2 some-value'.split())
        assert not mock.called

        # if we try passing an arbitrary unknown flag we explode
        # patching over the `error` method which usually sys.exit's
        # for any errors.
        parser.parse_args('a --level-1 some-value --some-flag'.split())
        assert mock.called

        patch_argument(parser, '--some-flag', action='store_true')
        mock.reset_mock()
        # now ever call combination accepts the flag we added
        assert parser.parse_args('--some-flag'.split())
        assert parser.parse_args('a --level-1 some-value --some-flag'.split())
        assert parser.parse_args('b --level-1 some-value --some-flag'.split())
        assert parser.parse_args('a a1 --level-2 some-value --some-flag'.split())
        assert parser.parse_args('b b1 --level-2 some-value --some-flag'.split())
        assert not mock.called




from gooey import Gooey, GooeyParser, constants


@Gooey(program_name="Gooey!!!!",
       tabbed_groups=False,
       naviation=constants.TABBED)
def parse_args():
    # import sys
    # print(sys.argv)
    parser = GooeyParser(description="Gooey turns your CLI apps into beautiful GUIs!")
    parser.add_argument('data_directory',
                        action='store',
                        default='scale=',
                        # type=lambda x: 'foo ' + x,
                        widget='DirChooser',
                        # gooey_options={
                        #     'validator': {
                        #         'test': "__import__('re').search('\d+x\d+', user_input) != None",
                        #         'message': 'oops'
                        #     }
                        # }
                        )
    import sys
    print(sys.argv)
    args = parser.parse_args()
    print(args)

if __name__ == '__main__':
    conf = parse_args()
    print("Done")


from typing import Optional, Tuple, List, Union, Mapping, Any, TypeVar, Generic, Dict

from dataclasses import dataclass
from typing_extensions import TypedDict


class MenuHtmlDialog(TypedDict):
    type: str
    menuTitle: str
    caption: Optional[str]
    html: str

class MenuLink(TypedDict):
    type: str
    menuTitle: str
    url: str


class MenuMessageDialog(TypedDict):
    type: str
    menuTitle: str
    message: str
    caption: Optional[str]

class MenuAboutDialog(TypedDict):
    type: str
    menuTitle: str
    name: Optional[str]
    description: Optional[str]
    version: Optional[str]
    copyright: Optional[str]
    license: Optional[str]
    website: Optional[str]
    developer: Optional[str]


MenuItem = Union[
    MenuLink,
    MenuMessageDialog,
    MenuAboutDialog,
    MenuHtmlDialog
]



class TimingOptions(TypedDict):
    show_time_remaining: bool
    hide_time_remaining_on_complete: bool


class GooeyParams(TypedDict):
    # when running with a custom target, there is no need to inject
    # --ignore-gooey into the CLI args
    show_preview_warning: bool
    suppress_gooey_flag: bool
    advanced: bool
    language: str
    target: Optional[str]
    program_name: Optional[str]
    program_description: Optional[str]
    sidebar_title: str
    default_size: Tuple[int, int]
    auto_start: bool
    show_advanced: bool
    run_validators: bool
    encoding: str
    show_stop_warning: bool
    show_success_modal: bool
    show_failure_modal: bool
    force_stop_is_error: bool
    poll_external_updates: bool  # BEING DEPRECATED
    return_to_config: bool
    show_restart_button: bool
    requires_shell: bool
    menu: List[MenuItem]
    clear_before_run: bool
    fullscreen: bool
    # Legacy/Backward compatibility interop
    use_legacy_titles: bool
    required_cols: int
    optional_cols: int
    manual_start: bool
    monospace_display: bool

    image_dir: str
    language_dir: str
    progress_regex: Optional[str]
    progress_expr: Optional[str]
    hide_progress_msg: bool
    timing_options: TimingOptions
    disable_progress_bar_animation: bool
    disable_stop_button: bool
    shutdown_signal: int
    use_events: List[str]

    # Layouts
    navigation: str
    show_sidebar: bool
    tabbed_groups: bool
    group_by_type: bool

    # styles
    body_bg_color: str
    header_bg_color: str
    header_height: int
    header_show_title: bool
    header_show_subtitle: bool
    header_image_center: bool
    footer_bg_color: str
    sidebar_bg_color: str

    # font family, weight, and size are determined at runtime
    terminal_panel_color: str
    terminal_font_color: str
    terminal_font_family: Optional[str]
    terminal_font_weight: Optional[int]
    terminal_font_size: Optional[int]
    richtext_controls: bool
    error_color: str

    use_cmd_args: bool
    dump_build_config: bool
    load_build_config: Optional[str]

# TODO:
# use the syntax here rather than inheritance, as the latter is a type error
# https://jdkandersson.com/2020/01/27/python-typeddict-arbitrary-key-names-with-totality/
# class BuildSpecification(GooeyParams):
#     target: str
#     widgets: str


class BasicField(TypedDict):
    id: str
    type: str
    # required: bool
    # positional: bool
    error: Optional[str]
    enabled: bool
    visible: bool

class Dropdown(BasicField):
    selected: int
    choices: List[str]

class Chooser(BasicField):
    btn_label: str
    value: str


class MultiFileChooser(Chooser):
    pass

class FileChooser(Chooser):
    pass

class FileSaver(Chooser):
    pass

class DirChooser(Chooser):
    pass

class MultiDirChooser(Chooser):
    pass

class DateChooser(Chooser):
    pass

class TimeChooser(Chooser):
    pass

class ColourChooser(Chooser):
    pass

class Command(BasicField):
    value: str
    placeholder: str

class Counter(BasicField):
    selected: int
    choices: List[str]

class DropdownFilterable(BasicField):
    value: str
    choices: List[str]

class Listbox(BasicField):
    selected: List[str]
    choices: List[str]

class IntegerField(BasicField):
    value: str
    min: int
    max: int

class DecimalField(BasicField):
    value: float
    min: float
    max: float

class Slider(BasicField):
    value: float
    min: float
    max: float

class Textarea(BasicField):
    value: float
    height: int

class TextField(BasicField):
    value: str
    placeholder: str


class PasswordField(TextField):
    pass


class Checkbox(BasicField):
    checked: bool


class RadioGroup(BasicField):
    selected: Optional[int]
    options: List['FormField']


FormField = Union[
    Textarea,
    Slider,
    Command,
    Counter,
    Checkbox,
    TextField,
    Dropdown,
    Chooser,
    RadioGroup,
    DropdownFilterable,
    Listbox,
    IntegerField
]




class FieldValue(TypedDict):
    """
    The current value of a widget in the UI.
    TODO: Why are things like cmd and cli type tracked IN the
    UI and returned as part of the getValue() call?
    What the hell, young me?
    """
    id: str
    cmd: Optional[str]
    rawValue: str
    placeholder: str
    positional: bool
    required: bool
    enabled: bool
    visible: bool
    test: bool
    error: Optional[str]
    clitype: str
    meta: Any


class PublicGooeyState(TypedDict):
    """
    A minimal representation of Gooey's current UI state
    """
    active_form: List[FormField]


class Group(TypedDict):
    name: str
    items: List['Item']
    groups: List['Group']
    description: str
    options: Dict[Any, Any]


class Item(TypedDict):
    id: str
    type: str
    cli_type: str
    group_name: str
    required: bool
    options: Dict[Any, Any]
    data: 'ItemData'


class EnrichedItem(Item):
    """
    An argparse item paired with its associated Gooey form
    field and current state.
    """
    field: FormField


ItemData = Union['StandardData', 'RadioData']

class StandardData(TypedDict):
    display_name: str
    help: str
    required: bool
    nargs: str
    commands: List[str]
    choices: List[str]
    default: Union[str, List[str]]
    dest: str

class RadioData(TypedDict):
    commands: List[List[str]]
    widgets: List[Item]


class TopLevelParser(TypedDict):
    command: str
    name: str
    help: Optional[str]
    description: str
    contents: List[Group]

A = TypeVar('A')


## TODO: dynamic types

@dataclass(frozen=True, eq=True)
class CommandDetails:
    target: str
    subcommand: str
    positionals: List[FieldValue]
    optionals: List[FieldValue]

@dataclass(frozen=True, eq=True)
class CommandPieces:
    target: str
    subcommand: str
    positionals: List[str]
    optionals: List[str]
    ignoreFlag: str

@dataclass(frozen=True, eq=True)
class Success(Generic[A]):
    value: A

    def map(self, f):
        return Success(f(self.value))
    def flatmap(self, f):
        return f(self.value)
    def onSuccess(self, f):
        f(self.value)
        return self
    def onError(self, f):
        return self
    def isSuccess(self):
        return True
    def getOrThrow(self):
        return self.value

@dataclass(frozen=True, eq=True)
class Failure:
    error: Exception

    def map(self, f):
        return Failure(self.error)
    def flatmap(self, f):
        return Failure(self.error)
    def onSuccess(self, f):
        return self
    def onError(self, f):
        f(self.error)
        return self
    def isSuccess(self):
        return False
    def getOrThrow(self):
        raise self.error

Try = Union[Success[A], Failure]



ValidationResponse = Mapping[str, str]


class InvalidChoiceException(ValueError):
    pass



import wx  # type: ignore

from gooey.gui.components.widgets.dropdown import Dropdown
from gooey.python_bindings import types as t
from gooey.gui import formatters


class Counter(Dropdown):

    def setValue(self, value):
        index = self._meta['choices'].index(value) + 1
        self.widget.SetSelection(index)

    def getUiState(self) -> t.FormField:
        widget: wx.ComboBox = self.widget
        return t.Counter(
            id=self._id,
            type=self.widgetInfo['type'],
            selected=self.getWidgetValue(),
            choices=widget.GetStrings(),
            error=self.error.GetLabel() or None,
            enabled=self.IsEnabled(),
            visible=self.IsShown()
        )



    def formatOutput(self, metadata, value):
        return formatters.counter(metadata, value)



import sys
import unittest
from copy import deepcopy

from gooey import Events
from gooey.tests.harness import instrumentGooey
from gooey.tests import *

class TestLiveDynamicUpdates(unittest.TestCase):

    def test_validate_form(self):
        """
        Integration testing the Dynamic Validation features.
        """
        # Because it's a live test, nothing is mocked. This basic.py file
        # will be called via subprocess as part of the test. As such, we
        # grab both its path on disk (for use as a target for Gooey) as
        # well as its parser instance (so that we can bootstrap)
        from gooey.tests.dynamics.files import basic
        params = {
            'target': '{} -u {}'.format(sys.executable, basic.__file__),
            'use_events': [Events.VALIDATE_FORM],
        }
        with instrumentGooey(basic.make_parser(), **params) as (app, frame, gapp):
            # the parser has a single arg of type int.
            # We purposefully give it invalid input for the sake of the test.
            gapp.getActiveConfig().widgetsMap['foo'].setValue('not a number')
            # and make sure we're not somehow starting with an error
            self.assertEqual(gapp.getActiveFormState()[0]['error'], '')
            gapp.onStart()

            # All subprocess calls ultimately pump though wx's event queue
            # so we have to kick off the mainloop and let it run long enough
            # to let the subprocess complete and the event queue flush
            wx.CallLater(2500, app.ExitMainLoop)
            app.MainLoop()
            # after the subprocess call is complete, our UI should have
            # been updated with the data dynamically returned from the
            # basic.py invocation.
            self.assertIn('invalid literal', gapp.getActiveFormState()[0]['error'])


    def test_validate_form_without_errors(self):
        from gooey.tests.dynamics.files import basic
        params = {
            'target': '{} -u {}'.format(sys.executable, basic.__file__),
            'use_events': [Events.VALIDATE_FORM],
            # setting to false because it interferes with the test
            'show_success_modal': False
        }
        with instrumentGooey(basic.make_parser(), **params) as (app, frame, gapp):
            gapp.getActiveConfig().widgetsMap['foo'].setValue('10')  # valid int
            self.assertEqual(gapp.getActiveFormState()[0]['error'], '')
            gapp.onStart()

            wx.CallLater(2500, app.ExitMainLoop)
            app.MainLoop()
            # no errors blocked the run, so we should have executed and finished.
            # we're now on the success screen.
            self.assertEqual(gapp.state['image'], gapp.state['images']['successIcon'])
            # and indeed no errors were written to the UI
            self.assertEqual(gapp.getActiveFormState()[0]['error'], '')
            # and we find the expected output written to the console
            # rather than some unexpected error
            self.assertIn('DONE', frame.FindWindowByName("console").getText())


    def test_lifecycle_handlers(self):
        cases = [
            {'input': 'happy path', 'expected_stdout': 'DONE', 'expected_update': 'success'},
            {'input': 'fail', 'expected_stdout': 'EXCEPTION', 'expected_update': 'error'}
        ]
        from gooey.tests.dynamics.files import lifecycles
        params = {
            'target': '{} -u {}'.format(sys.executable, lifecycles.__file__),
            'use_events': [Events.ON_SUCCESS, Events.ON_ERROR],
            'show_success_modal': False,
            'show_failure_modal': False
        }
        for case in cases:
            with self.subTest(case):
                with instrumentGooey(lifecycles.make_parser(), **params) as (app, frame, gapp):
                    gapp.getActiveConfig().widgetsMap['foo'].setValue(case['input'])
                    gapp.onStart()
                    # give everything a chance to run
                    wx.CallLater(2000, app.ExitMainLoop)
                    app.MainLoop()
                    # `lifecycle.py` is set up to raise an exception for certain inputs
                    # so we check that we find our expected stdout here
                    console = frame.FindWindowByName("console")
                    self.assertIn(case['expected_stdout'], console.getText())

                    # Now, based on what happened during the run (success/exception) our
                    # respective lifecycle handler should have been called. These are
                    # configured to update the form field in the UI with a relevant value.
                    # Thus we we're checking here to see that out input has changed, and now
                    # matches the value we expect from the handler
                    textfield = gapp.getActiveFormState()[0]
                    print(case['expected_update'], textfield['value'])
                    self.assertEqual(case['expected_update'], textfield['value'])




"""
Parser containing all Gooey widgets.
"""

from gooey import GooeyParser


parser = GooeyParser()

parser.add_argument('--textfield', default=2, widget="TextField")
parser.add_argument('--textarea', default="oneline twoline", widget='Textarea')
parser.add_argument('--password', default="hunter42", widget='PasswordField')
parser.add_argument('--commandfield', default="cmdr", widget='CommandField')
parser.add_argument('--dropdown', choices=["one", "two"], default="two", widget='Dropdown')
parser.add_argument(
    '--listboxie',
    nargs='+',
    default=['three', 'four'],
    choices=['one', 'two', 'three', 'four'],
    widget='Listbox',
    gooey_options={
        'height': 300,
        'validate': '',
        'heading_color': '',
        'text_color': '',
        'hide_heading': True,
        'hide_text': True,
    }
)
parser.add_argument('--counter', default=3, action='count', widget='Counter')
parser.add_argument("--overwrite1", action="store_true", default=True, widget='CheckBox')
parser.add_argument("--overwrite2", action="store_true", default=True, widget='BlockCheckbox')

verbosity = parser.add_mutually_exclusive_group(
    gooey_options={
        'initial_selection': 0
    }
)
verbosity.add_argument('--mutexone', default='hello')

parser.add_argument('--mutextwo', default='3', widget='Slider')
parser.add_argument('--mutextwo', default='1', widget='IntegerField')
parser.add_argument('--mutextwo', default='4', widget='DecimalField')

parser.add_argument("--filechooser", default="fc-value", widget='FileChooser')
parser.add_argument("--filesaver", default="fs-value", widget='FileSaver')
parser.add_argument("--dirchooser", default="dc-value", widget='DirChooser')
parser.add_argument("--datechooser", default="2015-01-01", widget='DateChooser')
parser.add_argument("--colourchooser", default="#000000", widget='ColourChooser')






"""
Primary orchestration and control point for Gooey.
"""
import queue
import sys
import threading
from contextlib import contextmanager
from functools import wraps
from json import JSONDecodeError
from pprint import pprint
from subprocess import CalledProcessError
from threading import Thread, get_ident
from typing import Mapping, Dict, Type, Iterable

import six
import wx  # type: ignore

from gooey.gui.state import FullGooeyState
from gooey.python_bindings.types import PublicGooeyState
from rewx.widgets import set_basic_props

from gooey.gui.components.mouse import notifyMouseEvent
from gooey.gui.state import initial_state, present_time, form_page, ProgressEvent, TimingEvent
from gooey.gui import state as s
from gooey.gui.three_to_four import Constants
from rewx.core import Component, Ref, updatewx, patch
from typing_extensions import TypedDict

from rewx import wsx, render, create_element, mount, update
from rewx import components as c
from wx.adv import TaskBarIcon  # type: ignore
import signal

from gooey import Events
from gooey.gui import cli
from gooey.gui import events
from gooey.gui import seeder
from gooey.gui.components import modals
from gooey.gui.components.config import ConfigPage, TabbedConfigPage
from gooey.gui.components.console import Console
from gooey.gui.components.footer import Footer
from gooey.gui.components.header import FrameHeader
from gooey.gui.components.menubar import MenuBar
from gooey.gui.components.sidebar import Sidebar
from gooey.gui.components.tabbar import Tabbar
from gooey.gui.lang.i18n import _
from gooey.gui.processor import ProcessController
from gooey.gui.util.time import Timing
from gooey.gui.pubsub import pub
from gooey.gui.util import wx_util
from gooey.gui.util.wx_util import transactUI
from gooey.python_bindings import constants
from gooey.python_bindings.types import Failure, Success, CommandDetails, Try
from gooey.util.functional import merge, associn, assoc
from gooey.gui.image_repository import loadImages
from gooey.gui import host


from threading import Lock

from gooey.util.functional import associnMany

lock = Lock()

class GooeyApplication(wx.Frame):
    """
    Main window for Gooey.
    """

    def __init__(self, buildSpec, *args, **kwargs):
        super(GooeyApplication, self).__init__(None, *args, **kwargs)
        self._state = {}
        self.buildSpec = buildSpec

        self.applyConfiguration()
        self.menu = MenuBar(buildSpec)
        self.SetMenuBar(self.menu)
        self.header = FrameHeader(self, buildSpec)
        self.configs = self.buildConfigPanels(self)
        self.navbar = self.buildNavigation()
        self.footer = Footer(self, buildSpec)
        self.console = Console(self, buildSpec)

        self.props = {
            'background_color': self.buildSpec['header_bg_color'],
            'title': self.buildSpec['program_name'],
            'subtitle': self.buildSpec['program_description'],
            'height': self.buildSpec['header_height'],
            'image_uri': self.buildSpec['images']['configIcon'],
            'image_size': (six.MAXSIZE, self.buildSpec['header_height'] - 10)}

        state = form_page(initial_state(self.buildSpec))

        self.fprops = {
            'buttons': state['buttons'],
            'progress': state['progress'],
            'timing': state['timing'],
            'bg_color': self.buildSpec['footer_bg_color']
        }

        # self.hhh = render(create_element(RHeader, self.props), self)
        # self.fff = render(create_element(RFooter, self.fprops), self)
        # patch(self.hhh, create_element(RHeader, {**self.props, 'image_uri': self.buildSpec['images']['runningIcon']}))
        self.layoutComponent()
        self.timer = Timing(self)

        self.clientRunner = ProcessController(
            self.buildSpec.get('progress_regex'),
            self.buildSpec.get('progress_expr'),
            self.buildSpec.get('hide_progress_msg'),
            self.buildSpec.get('encoding'),
            self.buildSpec.get('requires_shell'),
            self.buildSpec.get('shutdown_signal', signal.SIGTERM)
        )

        pub.subscribe(events.WINDOW_START, self.onStart)
        pub.subscribe(events.WINDOW_RESTART, self.onStart)
        pub.subscribe(events.WINDOW_STOP, self.onStopExecution)
        pub.subscribe(events.WINDOW_CLOSE, self.onClose)
        pub.subscribe(events.WINDOW_CANCEL, self.onCancel)
        pub.subscribe(events.WINDOW_EDIT, self.onEdit)
        pub.subscribe(events.CONSOLE_UPDATE, self.console.logOutput)
        pub.subscribe(events.EXECUTION_COMPLETE, self.onComplete)
        pub.subscribe(events.PROGRESS_UPDATE, self.footer.updateProgressBar)
        pub.subscribe(events.TIME_UPDATE, self.footer.updateTimeRemaining)
        # Top level wx close event
        # self.Bind(wx.EVT_CLOSE, self.onClose)

        # TODO: handle child focus for per-field level validation.
        # self.Bind(wx.EVT_CHILD_FOCUS, self.handleFocus)

        if self.buildSpec.get('auto_start', False):
            self.onStart()


    def applyConfiguration(self):
        self.SetTitle(self.buildSpec['program_name'])
        self.SetBackgroundColour(self.buildSpec.get('body_bg_color'))


    def onStart(self, *args, **kwarg):
        """
        Verify user input and kick off the client's program if valid
        """
        # navigates away from the button because a
        # disabled focused button still looks enabled.
        self.footer.cancel_button.Disable()
        self.footer.start_button.Disable()
        self.footer.start_button.Navigate()
        if Events.VALIDATE_FORM in self.buildSpec.get('use_events', []):
            # TODO: make this wx thread safe so that it can
            # actually run asynchronously
            Thread(target=self.onStartAsync).run()
        else:
            Thread(target=self.onStartAsync).run()

    def onStartAsync(self, *args, **kwargs):
        with transactUI(self):
            try:
                errors = self.validateForm().getOrThrow()
                if errors:  # TODO
                    config = self.navbar.getActiveConfig()
                    config.setErrors(errors)
                    self.Layout()
                    # TODO: account for tabbed layouts
                    # TODO: scroll the first error into view
                    # TODO: rather than just snapping to the top
                    self.configs[0].Scroll(0, 0)
                else:
                    if self.buildSpec['clear_before_run']:
                        self.console.clear()
                    self.clientRunner.run(self.buildCliString())
                    self.showConsole()
            except CalledProcessError as e:
                self.showError()
                self.console.appendText(str(e))
                self.console.appendText(
                    '\n\nThis failure happens when Gooey tries to invoke your '
                    'code for the VALIDATE_FORM event and receives an expected '
                    'error code in response.'
                )
                wx.CallAfter(modals.showFailure)
            except JSONDecodeError as e:
                self.showError()
                self.console.appendText(str(e))
                self.console.appendText(
                    '\n\nGooey was unable to parse the response to the VALIDATE_FORM event. '
                    'This can happen if you have additional logs to stdout beyond what Gooey '
                    'expects.'
                )
                wx.CallAfter(modals.showFailure)
            # for some reason, we have to delay the re-enabling of
            # the buttons by a few ms otherwise they pickup pending
            # events created while they were disabled. Trial and error
            # let to this solution.
            wx.CallLater(20, self.footer.start_button.Enable)
            wx.CallLater(20, self.footer.cancel_button.Enable)


    def onEdit(self):
        """Return the user to the settings screen for further editing"""
        with transactUI(self):
            for config in self.configs:
                config.resetErrors()
            self.showSettings()


    def onComplete(self, *args, **kwargs):
        """
        Display the appropriate screen based on the success/fail of the
        host program
        """
        with transactUI(self):
            if self.clientRunner.was_success():
                if self.buildSpec.get('return_to_config', False):
                    self.showSettings()
                else:
                    self.showSuccess()
                    if self.buildSpec.get('show_success_modal', True):
                        wx.CallAfter(modals.showSuccess)
            else:
                if self.clientRunner.wasForcefullyStopped:
                    self.showForceStopped()
                else:
                    self.showError()
                    if self.buildSpec.get('show_failure_modal'):
                        wx.CallAfter(modals.showFailure)

    def onCancel(self):
        """Close the program after confirming

        We treat the behavior of the "cancel" button slightly
        differently than the general window close X button only
        because this is 'part of' the form.
        """
        if modals.confirmExit():
            self.onClose()


    def onStopExecution(self):
        """Displays a scary message and then force-quits the executing
        client code if the user accepts"""
        if self.shouldStopExecution():
            self.clientRunner.stop()


    def onClose(self, *args, **kwargs):
        """Stop any actively running client program, cleanup the top
        level WxFrame and shutdown the current process"""
        # issue #592 - we need to run the same onStopExecution machinery
        # when the exit button is clicked to ensure everything is cleaned
        # up correctly.
        if self.clientRunner.running():
            if self.shouldStopExecution():
                self.clientRunner.stop()
                self.destroyGooey()
        else:
            self.destroyGooey()

    def buildCliString(self) -> str:
        """
        Collect all of the required information from the config screen and
        build a CLI string which can be used to invoke the client program
        """
        cmd = self.getCommandDetails()
        return cli.cliCmd(
            cmd.target,
            cmd.subcommand,
            cmd.positionals,
            cmd.optionals,
            suppress_gooey_flag=self.buildSpec['suppress_gooey_flag']
        )

    def validateForm(self) -> Try[Mapping[str, str]]:
        config = self.navbar.getActiveConfig()
        localErrors: Mapping[str, str] = config.getErrors()
        dynamicResult: Try[Mapping[str, str]] = self.fetchDynamicValidations()

        combineErrors = lambda m: merge(localErrors, m)
        return dynamicResult.map(combineErrors)


    def fetchDynamicValidations(self) -> Try[Mapping[str, str]]:
        # only run the dynamic validation if the user has
        # specifically subscribed to that event
        if Events.VALIDATE_FORM in self.buildSpec.get('use_events', []):
            cmd = self.getCommandDetails()
            return seeder.communicate(cli.formValidationCmd(
                cmd.target,
                cmd.subcommand,
                cmd.positionals,
                cmd.optionals
            ), self.buildSpec['encoding'])
        else:
            # shim response if nothing to do.
            return Success({})


    def getCommandDetails(self) -> CommandDetails:
        """
        Temporary helper for getting the state of the current Config.

        To be deprecated upon (the desperately needed) refactor.
        """
        config = self.navbar.getActiveConfig()
        group = self.buildSpec['widgets'][self.navbar.getSelectedGroup()]
        return CommandDetails(
            self.buildSpec['target'],
            group['command'],
            config.getPositionalValues(),
            config.getOptionalValues(),
        )


    def shouldStopExecution(self):
        return not self.buildSpec['show_stop_warning'] or modals.confirmForceStop()


    def destroyGooey(self):
        self.Destroy()
        sys.exit()

    def block(self, **kwargs):
        pass


    def layoutComponent(self):
        sizer = wx.BoxSizer(wx.VERTICAL)
        # sizer.Add(self.hhh, 0, wx.EXPAND)
        sizer.Add(self.header, 0, wx.EXPAND)
        sizer.Add(wx_util.horizontal_rule(self), 0, wx.EXPAND)

        sizer.Add(self.navbar, 1, wx.EXPAND)
        sizer.Add(self.console, 1, wx.EXPAND)
        sizer.Add(wx_util.horizontal_rule(self), 0, wx.EXPAND)
        # sizer.Add(self.fff, 0, wx.EXPAND)
        sizer.Add(self.footer, 0, wx.EXPAND)
        self.SetMinSize((400, 300))
        self.SetSize(self.buildSpec['default_size'])
        self.SetSizer(sizer)
        self.console.Hide()
        self.Layout()
        if self.buildSpec.get('fullscreen', True):
            self.ShowFullScreen(True)
        # Program Icon (Windows)
        icon = wx.Icon(self.buildSpec['images']['programIcon'], wx.BITMAP_TYPE_PNG)
        self.SetIcon(icon)
        if sys.platform != 'win32':
            # OSX needs to have its taskbar icon explicitly set
            # bizarrely, wx requires the TaskBarIcon to be attached to the Frame
            # as instance data (self.). Otherwise, it will not render correctly.
            self.taskbarIcon = TaskBarIcon(iconType=wx.adv.TBI_DOCK)
            self.taskbarIcon.SetIcon(icon)


    def buildNavigation(self):
        """
        Chooses the appropriate layout navigation component based on user prefs
        """
        if self.buildSpec['navigation'] == constants.TABBED:
            navigation = Tabbar(self, self.buildSpec, self.configs)
        else:
            navigation = Sidebar(self, self.buildSpec, self.configs)
            if self.buildSpec['navigation'] == constants.HIDDEN:
                navigation.Hide()
        return navigation


    def buildConfigPanels(self, parent):
        page_class = TabbedConfigPage if self.buildSpec['tabbed_groups'] else ConfigPage

        return [page_class(parent, widgets, self.buildSpec)
                for widgets in self.buildSpec['widgets'].values()]


    def showSettings(self):
        self.navbar.Show(True)
        self.console.Show(False)
        self.header.setImage('settings_img')
        self.header.setTitle(_("settings_title"))
        self.header.setSubtitle(self.buildSpec['program_description'])
        self.footer.showButtons('cancel_button', 'start_button')
        self.footer.progress_bar.Show(False)
        self.footer.time_remaining_text.Show(False)


    def showConsole(self):
        self.navbar.Show(False)
        self.console.Show(True)
        self.header.setImage('running_img')
        self.header.setTitle(_("running_title"))
        self.header.setSubtitle(_('running_msg'))
        self.footer.showButtons('stop_button')
        if not self.buildSpec.get('disable_progress_bar_animation', False):
            self.footer.progress_bar.Show(True)
        self.footer.time_remaining_text.Show(False)
        if self.buildSpec.get('timing_options')['show_time_remaining']:
            self.timer.start()
            self.footer.time_remaining_text.Show(True)
        if not self.buildSpec['progress_regex']:
            self.footer.progress_bar.Pulse()


    def showComplete(self):
        self.navbar.Show(False)
        self.console.Show(True)
        buttons = (['edit_button', 'restart_button', 'close_button']
                   if self.buildSpec.get('show_restart_button', True)
                   else ['edit_button', 'close_button'])
        self.footer.showButtons(*buttons)
        self.footer.progress_bar.Show(False)
        if self.buildSpec.get('timing_options')['show_time_remaining']:
            self.timer.stop()
        self.footer.time_remaining_text.Show(True)
        if self.buildSpec.get('timing_options')['hide_time_remaining_on_complete']:
            self.footer.time_remaining_text.Show(False)


    def showSuccess(self):
        self.showComplete()
        self.header.setImage('check_mark')
        self.header.setTitle(_('finished_title'))
        self.header.setSubtitle(_('finished_msg'))
        self.Layout()


    def showError(self):
        self.showComplete()
        self.header.setImage('error_symbol')
        self.header.setTitle(_('finished_title'))
        self.header.setSubtitle(_('finished_error'))


    def showForceStopped(self):
        self.showComplete()
        if self.buildSpec.get('force_stop_is_error', True):
            self.showError()
        else:
            self.showSuccess()
        self.header.setSubtitle(_('finished_forced_quit'))









import re
import signal
import subprocess
import sys
import unittest
import os
import time

import wx

from gooey.gui import events, processor
from gooey.gui.pubsub import pub
from gooey.gui.processor import ProcessController


class TestProcessor(unittest.TestCase):

    def test_extract_progress(self):
        # should pull out a number based on the supplied
        # regex and expression
        processor = ProcessController(r"^progress: (\d+)%$", None, False, 'utf-8')
        self.assertEqual(processor._extract_progress(b'progress: 50%'), 50)

        processor = ProcessController(r"total: (\d+)%$", None, False, 'utf-8')
        self.assertEqual(processor._extract_progress(b'my cool total: 100%'), 100)

    def test_extract_progress_returns_none_if_no_regex_supplied(self):
        processor = ProcessController(None, None, False, 'utf-8')
        self.assertIsNone(processor._extract_progress(b'Total progress: 100%'))


    def test_extract_progress_returns_none_if_no_match_found(self):
        processor = ProcessController(r'(\d+)%$', None, False, 'utf-8')
        self.assertIsNone(processor._extract_progress(b'No match in dis string'))


    def test_eval_progress(self):
        # given a match in the string, should eval the result
        regex = r'(\d+)/(\d+)$'
        processor = ProcessController(regex, r'x[0] / x[1]', False,False, 'utf-8')
        match = re.search(regex, '50/50')
        self.assertEqual(processor._eval_progress(match), 1.0)
    def test_eval_progress_returns_none_on_failure(self):
        # given a match in the string, should eval the result
        regex = r'(\d+)/(\d+)$'
        processor = ProcessController(regex, r'x[0] *^/* x[1]', False, False,'utf-8')
        match = re.search(regex, '50/50')
        self.assertIsNone(processor._eval_progress(match))


    def test_all_interrupts_halt_process(self):
        """
        TODO: These tests are hella flaky. I'm confident that the feature works. However, getting
        signals, subprocesses and unittest to all play together reliably is proving tricky. It
        primarily seems to come down to how long the time.sleep() is before sending the shutdown
        signal.
        """

        cmd = 'python ' + os.path.join(os.getcwd(), 'files', 'infinite_loop.py')

        try:
            import _winapi
            signals = [signal.SIGTERM, signal.CTRL_BREAK_EVENT, signal.CTRL_C_EVENT]
        except ModuleNotFoundError:
            signals = [signal.SIGTERM, signal.SIGINT]
        try:
            for sig in signals:
                print('sig', sig)
                processor = ProcessController(None, None, False, 'utf-8', True, shutdown_signal=sig)

                processor.run(cmd)
                self.assertTrue(processor.running())

                # super-duper important sleep so that the
                # signal is actually received by the child process
                # see: https://stackoverflow.com/questions/32023719/how-to-simulate-a-terminal-ctrl-c-event-from-a-unittest
                time.sleep(1)
                processor.stop()
                max_wait = time.time() + 4
                while processor.running() and time.time() < max_wait:
                    time.sleep(0.1)
                self.assertFalse(processor.running())
        except KeyboardInterrupt:
            pass


    def test_ignore_sigint_family_signals(self):
        try:
            import _winapi
            signals = [signal.CTRL_BREAK_EVENT, signal.CTRL_C_EVENT]
            programs = ['ignore_break.py', 'ignore_interrupt.py']
        except ModuleNotFoundError:
            signals = [signal.SIGINT]
            programs = ['ignore_interrupt.py']


        for program, sig in zip(programs, signals):
            cmd = sys.executable + ' ' + os.path.join(os.getcwd(), 'files', program)
            process = processor = ProcessController(None, None, False, 'utf-8', True, shutdown_signal=sig, testmode=True)
            process.run(cmd)
            # super-duper important sleep so that the
            # signal is actually received by the child process
            # see: https://stackoverflow.com/questions/32023719/how-to-simulate-a-terminal-ctrl-c-event-from-a-unittest
            time.sleep(1)
            process.send_shutdown_signal()
            # wait to give stdout enough time to write
            time.sleep(1)
            # now our signal should have been received, but rejected.
            self.assertTrue(processor.running())
            # so we sigterm to actually shut down the process.
            process._send_signal(signal.SIGTERM)
            # sanity wait
            max_wait = time.time() + 2
            while processor.running() and time.time() < max_wait:
                time.sleep(0.1)
            # now we should be shut down due to killing the process.
            self.assertFalse(processor.running())
            # and we'll see in the stdout out from the process that our
            # interrupt was received
            output = process._process.stdout.read().decode('utf-8')
            self.assertIn("INTERRUPT", str(output))
            # but indeed ignored. It continued running and writing to stdout after
            # receiving the signal
            self.assertTrue(output.index("INTERRUPT") < len(output))




import signal
import sys
import textwrap

import os
from typing import List

from gooey.python_bindings.constants import Events
from gooey.python_bindings import constants
from gooey.gui.util.freeze import getResourcePath
from gooey.python_bindings.types import GooeyParams
from gooey.util.functional import merge



def _get_font_weight(kwargs):
    error_msg = textwrap.dedent('''
    Unknown font weight {}. 

    The available weights can be found in the `constants` module. 
    They're prefixed with "FONTWEIGHT" (e.g. `FONTWEIGHT_BOLD`)

    example code:    

    ```
    from gooey import constants
    @Gooey(terminal_font_weight=constants.FONTWEIGHT_NORMAL)
    ```   
    ''')
    weights = {
        constants.FONTWEIGHT_THIN,
        constants.FONTWEIGHT_EXTRALIGHT,
        constants.FONTWEIGHT_LIGHT,
        constants.FONTWEIGHT_NORMAL,
        constants.FONTWEIGHT_MEDIUM,
        constants.FONTWEIGHT_SEMIBOLD,
        constants.FONTWEIGHT_BOLD,
        constants.FONTWEIGHT_EXTRABOLD,
        constants.FONTWEIGHT_HEAVY,
        constants.FONTWEIGHT_EXTRAHEAVY
    }
    weight = kwargs.get('terminal_font_weight', constants.FONTWEIGHT_NORMAL)
    if weight not in weights:
        raise ValueError(error_msg.format(weight))
    return weight


# python can't type kwargs? wtf..
def gooey_params(**kwargs) -> GooeyParams:
    """
    Builds the full GooeyParams object from an arbitrary subset of supplied values
    """
    return GooeyParams(**{  # type: ignore
        'show_preview_warning': kwargs.get('show_preview_warning', True),
        'language': kwargs.get('language', 'english'),
        'target': kwargs.get('target'),

        'dump_build_config': kwargs.get('dump_build_config', False),
        'load_build_config': kwargs.get('load_build_config'),
        'use_cmd_args': kwargs.get('use_cmd_args', False),

        'suppress_gooey_flag': kwargs.get('suppress_gooey_flag') or False,
        # TODO: I should not read from the environment.
        # remains here for legacy reasons pending refactor
        'program_name': kwargs.get('program_name') or os.path.basename(sys.argv[0]).replace('.py', ''),
        'program_description': kwargs.get('program_description') or '',
        'sidebar_title': kwargs.get('sidebar_title', 'Actions'),
        'default_size': kwargs.get('default_size', (610, 530)),
        'auto_start': kwargs.get('auto_start', False),
        'advanced': kwargs.get('advanced', True),
        'run_validators': kwargs.get('run_validators', True),
        'encoding': kwargs.get('encoding', 'utf-8'),
        'show_stop_warning': kwargs.get('show_stop_warning', True),
        'show_success_modal': kwargs.get('show_success_modal', True),
        'show_failure_modal': kwargs.get('show_failure_modal', True),
        'force_stop_is_error': kwargs.get('force_stop_is_error', True),
        'poll_external_updates': kwargs.get('poll_external_updates', False),
        'return_to_config': kwargs.get('return_to_config', False),
        'show_restart_button': kwargs.get('show_restart_button', True),
        'requires_shell': kwargs.get('requires_shell', True),
        'menu': kwargs.get('menu', []),
        'clear_before_run': kwargs.get('clear_before_run', False),
        'fullscreen': kwargs.get('fullscreen', False),

        'use_legacy_titles': kwargs.get('use_legacy_titles', True),
        'required_cols': kwargs.get('required_cols', 2),
        'optional_cols': kwargs.get('optional_cols', 2),
        'manual_start': False,
        'monospace_display': kwargs.get('monospace_display', False),

        'image_dir': kwargs.get('image_dir', '::gooey/default'),
        # TODO: this directory resolution shouldn't happen here!
        # TODO: leaving due to legacy for now
        'language_dir': kwargs.get('language_dir', getResourcePath('languages')),
        'progress_regex': kwargs.get('progress_regex'),
        'progress_expr': kwargs.get('progress_expr'),
        'hide_progress_msg': kwargs.get('hide_progress_msg', False),

        'timing_options': merge({
            'show_time_remaining': False,
            'hide_time_remaining_on_complete': True
        }, kwargs.get('timing_options', {})),
        'disable_progress_bar_animation': kwargs.get('disable_progress_bar_animation', False),
        'disable_stop_button': kwargs.get('disable_stop_button'),
        'shutdown_signal': kwargs.get('shutdown_signal', signal.SIGTERM),
        'use_events': parse_events(kwargs.get('use_events', [])),


        'navigation': kwargs.get('navigation', constants.SIDEBAR),
        'show_sidebar': kwargs.get('show_sidebar', False),
        'tabbed_groups': kwargs.get('tabbed_groups', False),
        'group_by_type': kwargs.get('group_by_type', True),


        'body_bg_color': kwargs.get('body_bg_color', '#f0f0f0'),
        'header_bg_color': kwargs.get('header_bg_color', '#ffffff'),
        'header_height': kwargs.get('header_height', 90),
        'header_show_title': kwargs.get('header_show_title', True),
        'header_show_subtitle': kwargs.get('header_show_subtitle', True),
        'header_image_center': kwargs.get('header_image_center', False),
        'footer_bg_color': kwargs.get('footer_bg_color', '#f0f0f0'),
        'sidebar_bg_color': kwargs.get('sidebar_bg_color', '#f2f2f2'),

        'terminal_panel_color': kwargs.get('terminal_panel_color', '#F0F0F0'),
        'terminal_font_color': kwargs.get('terminal_font_color', '#000000'),
        'terminal_font_family': kwargs.get('terminal_font_family', None),
        'terminal_font_weight': _get_font_weight(kwargs),
        'terminal_font_size': kwargs.get('terminal_font_size', None),
        'richtext_controls': kwargs.get('richtext_controls', False),
        'error_color': kwargs.get('error_color', '#ea7878'),
        # TODO: remove. Only useful for testing
        'cli': kwargs.get('cli', sys.argv),
    })


def parse_events(events: List[str]) -> List[str]:
    if not isinstance(events, list):
        raise TypeError(
            f"`use_events` requires a list of events. You provided "
            "{events}. \n"
            "Example: \n"
            "\tfrom gooey import Events"
            "\t@Gooey(use_events=[Events.VALIDATE_FORM]")

    unknown_events = set(events) - set(Events)
    if unknown_events:
        raise ValueError(
            f'nrecognized event(s) were passed to `use_events`: {unknown_events}\n'
            f'Must be one of {Events._fields}\n'
            f'Consider using the `Events` object: `from gooey import Events`')
    else:
        return events



"""
Created on Jan 24, 2014  <-- so long ago!
"""
import sys
from argparse import ArgumentParser
from functools import wraps

from gooey.python_bindings.control import choose_hander
from gooey.python_bindings.parameters import gooey_params
from gooey.python_bindings.types import GooeyParams

IGNORE_COMMAND = '--ignore-gooey'


def Gooey(f=None, **gkwargs):
    """
    Decoration entry point for the Gooey process.
    See types.GooeyParams for kwargs options
    """
    params: GooeyParams = gooey_params(**gkwargs)

    @wraps(f)
    def inner(*args, **kwargs):
        parser_handler = choose_hander(params, gkwargs.get('cli', sys.argv))
        # monkey patch parser
        ArgumentParser.original_parse_args = ArgumentParser.parse_args
        ArgumentParser.parse_args = parser_handler
        # return the wrapped, now monkey-patched, user function
        # to be later invoked
        return f(*args, **kwargs)

    def thunk(func):
        """
        This just handles the case where the decorator is called
        with arguments (i.e. @Gooey(foo=bar) rather than @Gooey).

        Cause python is weird, when a decorator is called (e.g. @decorator())
        rather than just declared (e.g. @decorator), in complete and utter
        defiance of what your lying eyes see, it changes from a higher order
        function, to a function that takes an arbitrary argument *and then*
        returns a higher order function. i.e.

        decorate :: (a -> b) -> (a -> b)
        decorate() :: c -> (a -> b) -> (a -> b)

        wat.
        """
        return Gooey(func, **params)

    return inner if callable(f) else thunk




import wx  # type: ignore

from gooey.gui import formatters
from gooey.gui.components.widgets.bases import TextContainer
from gooey.python_bindings import types as t


class Listbox(TextContainer):

    def getWidget(self, parent, *args, **options):
        height = self._options.get('height', 60)
        return wx.ListBox(
            parent=parent,
            choices=self._meta['choices'],
            size=(-1, height),
            style=wx.LB_MULTIPLE
        )

    def setOptions(self, options):
        self.widget.Clear()
        for option in options:
            self.widget.Append(option)

    def setValue(self, values):
        for string in values:
            self.widget.SetStringSelection(string)

    def getWidgetValue(self):
        return [self.widget.GetString(index)
                for index in self.widget.GetSelections()]

    def formatOutput(self, metadata, value):
        return formatters.listbox(metadata, value)

    def getUiState(self) -> t.FormField:
        widget: wx.ComboBox = self.widget
        return t.Listbox(
            id=self._id,
            type=self.widgetInfo['type'],
            selected=self.getWidgetValue(),
            choices=self._meta['choices'],
            error=self.error.GetLabel() or None,
            enabled=self.IsEnabled(),
            visible=self.IsShown()
        )

    def syncUiState(self, state: t.Listbox):  # type: ignore
        widget: wx.ComboBox = self.widget
        widget.Clear()
        widget.AppendItems(state.get('choices', []))
        for string in state['selected']:
            widget.SetStringSelection(string)
        self.error.SetLabel(state['error'] or '')
        self.error.Show(state['error'] is not None and state['error'] is not '')



"""
Util for talking to the client program in order to retrieve
dynamic defaults for the UI
"""
import subprocess
from json import JSONDecodeError
from subprocess import CalledProcessError

from gooey.python_bindings.types import Try, Success, Failure
from gooey.python_bindings.coms import deserialize_inbound


def communicate(cmd, encoding) -> Try:
    """
    Invoke the processes specified by `cmd`.
    Assumes that the process speaks JSON over stdout. Non-json response
    are treated as an error.

    Implementation Note: I don't know why, but `Popen` is like ~5-6x faster
    than `check_output`. in practice, it means waiting for ~1/10th
    of a second rather than ~7/10ths of a second. A
    difference which is pretty weighty when there's a
    user waiting on the other end.
    """
    try:
        proc = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )
        out, err = proc.communicate()
        if out and proc.poll() == 0:
            return Success(deserialize_inbound(out, encoding))
        else:
            return Failure(CalledProcessError(proc.returncode, cmd, output=out, stderr=err))
    except JSONDecodeError as e:
        return Failure(e)




"""
Python file for Processor test suite

Infinite loop which would continue forever if not
interrupted.
"""
import time
import sys

if sys.platform.startswith('win'):
    import ctypes
    kernel32 = ctypes.WinDLL('kernel32')
    kernel32.SetConsoleCtrlHandler(None, 0)


while True:
    print(time.time())
    time.sleep(0.1)


from typing import Dict, Any

from gooey.python_bindings.types import PublicGooeyState
from gooey.python_bindings import types as t


def validate_public_state(state: Dict[str, Any]) -> PublicGooeyState:
    """
    Very, very minimal validation the shape of the incoming state
    is inline with the PublicGooeyState type.

    TODO: turn this into something useful.
    """
    top_level_keys = PublicGooeyState.__annotations__.keys()
    assert set(top_level_keys) == set(state.keys())
    for item in state['active_form']:
        assert 'type' in item
        expected_keys = getattr(t, item['type']).__annotations__.keys()
        a = set(expected_keys)
        b = set(item.keys())
        assert set(expected_keys) == set(item.keys())
    return state



from typing import Mapping

from gooey import Events
from gooey.python_bindings.types import Try
from gooey.util.functional import merge


def validateForm(self) -> Try[Mapping[str, str]]:  # or Exception
    config = self.navbar.getActiveConfig()
    localErrors: Mapping[str, str] = config.getErrors()
    dynamicResult: Try[Mapping[str, str]] = self.fetchDynamicValidations()

    combineErrors = lambda m: merge(localErrors, m)
    return dynamicResult.map(combineErrors)


def fetchDynamicValidations(self) -> Try[Mapping[str, str]]:
    # only run the dynamic validation if the user has
    # specifically subscribed to that event
    if Events.VALIDATE_FORM in self.buildSpec.get('use_events', []):
        cmd = self.getCommandDetails()
        return seeder.communicate2(cli.formValidationCmd(
            cmd.target,
            cmd.subcommand + 'baba',
            cmd.positionals,
            cmd.optionals
        ), self.buildSpec['encoding'])
    else:
        # shim response if nothing to do.
        return Success({})


"""
Houses all the supporting rewx components for
the main application window.
"""
import wx  # type: ignore
from typing_extensions import TypedDict

from gooey.gui.components.config import ConfigPage, TabbedConfigPage
from gooey.gui.components.console import Console
from gooey.gui.components.mouse import notifyMouseEvent
from gooey.gui.components.sidebar import Sidebar
from gooey.gui.components.tabbar import Tabbar
from gooey.gui.lang.i18n import _
from gooey.gui.pubsub import pub
from gooey.gui.state import present_time
from gooey.gui.three_to_four import Constants
from gooey.python_bindings import constants
from rewx import components as c  # type: ignore
from rewx import wsx, mount, update  # type: ignore
from rewx.core import Component, Ref  # type: ignore
from rewx.widgets import set_basic_props  # type: ignore


def attach_notifier(parent):
    """
    Recursively attaches the mouseEvent notifier
    to all elements in the tree
    """
    parent.Bind(wx.EVT_LEFT_DOWN, notifyMouseEvent)
    for child in parent.Children:
        attach_notifier(child)


class HeaderProps(TypedDict):
    background_color: str
    title: str
    show_title: bool
    subtitle: str
    show_subtitle: bool


class RHeader(Component):
    def __init__(self, props):
        super().__init__(props)
        self.parentRef = Ref()

    def component_did_mount(self):
        attach_notifier(self.parentRef.instance)

    def render(self):
        if 'running' not in self.props['image_uri']:
            imageProps = {
                'uri': self.props['image_uri'],
                'size': self.props['image_size'],
                'flag': wx.RIGHT,
                'border': 10}
        else:
            imageProps = {
                'size': self.props['image_size'],
                'flag': wx.RIGHT,
                'border': 10}
        return wsx(
            [c.Block, {'orient': wx.HORIZONTAL,
                       'ref': self.parentRef,
                       'min_size': (120, self.props['height']),
                       'background_color': self.props['background_color']},
             [c.Block, {'orient': wx.VERTICAL,
                        'flag': wx.ALIGN_CENTER_VERTICAL | wx.ALL,
                        'proportion': 1,
                        'border': 10},
              [TitleText, {'label': self.props['title'],
                           'show': self.props['show_title'],
                           'wx_name': 'header_title'}],
              [c.StaticText, {'label': self.props['subtitle'],
                              'show': self.props['show_subtitle'],
                              'wx_name': 'header_subtitle'}]],
             [c.StaticBitmap, imageProps]]
        )



class RFooter(Component):
    def __init__(self, props):
        super().__init__(props)
        self.ref = Ref()

    def component_did_mount(self):
        """
        We have to manually wire up LEFT_DOWN handlers
        for every component due to wx limitations.
        See: mouse.py docs for background.
        """
        block: wx.BoxSizer = self.ref.instance
        attach_notifier(block)

    def handle(self, btn):
        def inner(*args, **kwargs):
            pub.send_message(btn['id'])
        return inner

    def render(self):
        return wsx(
            [c.Block, {'orient': wx.VERTICAL,
                       'min_size': (30, 53),
                       'background_color': self.props['bg_color']},
             [c.Block, {'orient': wx.VERTICAL, 'proportion': 1}],
             [c.Block, {'orient': wx.HORIZONTAL,
                        'border': 20,
                        'flag': wx.EXPAND | wx.LEFT | wx.RIGHT,
                        'ref': self.ref},
              [c.Gauge, {'range': 100,
                         'proportion': 1,
                         'value': self.props['progress']['value'],
                         'show': self.props['progress']['show']}],
              [c.StaticText, {'label': present_time(self.props['timing']),
                              'flag': wx.LEFT,
                              'wx_name': 'timing',
                              'show': self.props['timing']['show'],
                              'border': 20}],
              [c.Block, {'orient': wx.HORIZONTAL, 'proportion': 1}],
              *[[c.Button, {**btn,
                            'label': _(btn['label_id']),
                            'min_size': (90, 23),
                            'flag': wx.LEFT,
                            'border': 10,
                            'on_click': self.handle(btn)
                            }]
                for btn in self.props['buttons']]],
             [c.Block, {'orient': wx.VERTICAL, 'proportion': 1}]]
        )


class RNavbar(Component):
    def __init__(self, props):
        super().__init__(props)

    # if self.buildSpec['navigation'] == constants.TABBED:
    #     navigation = Tabbar(self, self.buildSpec, self.configs)
    # else:
    #     navigation = Sidebar(self, self.buildSpec, self.configs)
    #     if self.buildSpec['navigation'] == constants.HIDDEN:
    #         navigation.Hide()
    def render(self):
        return wsx(

        )

def VerticalSpacer(props):
    return wsx([c.Block, {'orient': wx.VERTICAL, 'min_size': (-1, props['height'])}])

def SidebarControls(props):
    return wsx(
        [c.Block, {'orient': wx.VERTICAL,
                   'min_size': (180, 0),
                   'size': (180, 0),
                   'show': props.get('show', True),
                   'flag': wx.EXPAND,
                   'proportion': 0,
                   'background_color': props['bg_color']},
         [c.Block, {'orient': wx.VERTICAL,
                    'min_size': (180, 0),
                    'size': (180, 0),
                    'flag': wx.EXPAND | wx.LEFT | wx.RIGHT | wx.BOTTOM,
                    'border': 10,
                    'proportion': 1,
                    'background_color': props['bg_color']},
          [VerticalSpacer, {'height': 15}],
          [TitleText, {'label': props['label']}],
          [VerticalSpacer, {'height': 5}],
          [c.ListBox, {'choices': props['options'],
                       'value': props['activeSelection'],
                       'proportion': 1,
                       'on_change': props['on_change'],
                       'flag': wx.EXPAND}],
          [VerticalSpacer, {'height': 10}]]]
    )


def ProgressSpinner(props):
    return wsx(
        [c.Block, {'flag': wx.EXPAND, 'show': props['show']},
         [c.Gauge, {'flag': wx.EXPAND,
                    'value': -1,
                    'size': (-1, 4)}],
         [c.StaticLine, {'style': wx.LI_HORIZONTAL,
                         'flag': wx.EXPAND}]]
    )


def ErrorWarning(props):
    return wsx(
        [c.Block, {'orient': wx.HORIZONTAL,
                   'background_color': '#fdeded',
                   'style': wx.SIMPLE_BORDER,
                   'flag': wx.EXPAND | wx.ALL,
                   'proportion': 0,
                   'border': 5,
                   'min_size': (-1, 45),
                   'show': props.get('show', True)},
         [c.StaticBitmap, {'size': (24, 24),
                           'flag': wx.LEFT | wx.RIGHT | wx.ALIGN_CENTER_VERTICAL,
                           'border': 6,
                           'uri': props['uri']}],
         [c.StaticText, {'label': 'Whoops! You have some errors which must be corrected',
                         'flag': wx.ALIGN_CENTER_VERTICAL}]]
    )

def RSidebar(props):
    return wsx(
        [c.Block, {'orient': wx.HORIZONTAL,
                   'show': props.get('show', True),
                   'flag': props['flag'],
                   'proportion': props['proportion'],
                   'ref': props['ref']},
         [SidebarControls, {**props, 'show': props['show_sidebar']}],
         [c.StaticLine, {'style': wx.LI_VERTICAL,
                         'flag': wx.EXPAND,
                         'min_size': (1, -1)}],
         *[[TabbedConfigPage if props['tabbed_groups'] else ConfigPage,
            {'flag': wx.EXPAND,
             'proportion': 3,
             'config': config,
             'show': i == props['activeSelection']}]
           for i, config in enumerate(props['config'].values())]
         ]
    )


def RTabbedLayout(props):
    return wsx(
        [c.Notebook, {'flag': wx.EXPAND | wx.ALL,
                      'show': props.get('show', True),
                      'proportion': 1,
                      'on_change': props['on_change'],
                      'ref': props['ref']},
         *[[c.NotebookItem,
            {'title': props['options'][i], 'selected': props['activeSelection'] == i},
            [TabbedConfigPage if props['tabbed_groups'] else ConfigPage,
             {'flag': wx.EXPAND,
              'proportion': 3,
              'config': config,
              'show': i == props['activeSelection']}]]
           for i, config in enumerate(props['config'].values())]]
    )



def layout_choose():
    def buildNavigation(self):
        """
        Chooses the appropriate layout navigation component based on user prefs
        """
        if self.buildSpec['navigation'] == constants.TABBED:
            navigation = Tabbar(self, self.buildSpec, self.configs)
        else:
            navigation = Sidebar(self, self.buildSpec, self.configs)
            if self.buildSpec['navigation'] == constants.HIDDEN:
                navigation.Hide()
        return navigation


    def buildConfigPanels(self, parent):
        page_class = TabbedConfigPage if self.buildSpec['tabbed_groups'] else ConfigPage

        return [page_class(parent, widgets, self.buildSpec)
                for widgets in self.buildSpec['widgets'].values()]















class TitleText(Component):
    def __init__(self, props):
        super().__init__(props)
        self.ref = Ref()

    def component_did_mount(self):
        text: wx.StaticText = self.ref.instance
        font_size = text.GetFont().GetPointSize()
        text.SetFont(wx.Font(
            int(font_size * 1.2),
            wx.FONTFAMILY_DEFAULT,
            Constants.WX_FONTSTYLE_NORMAL,
            wx.FONTWEIGHT_BOLD,
            False
        ))

    def render(self):
        return wsx([c.StaticText, {**self.props, 'label': self.props['label'], 'ref': self.ref}])


##
## REWX definitions:
##

@mount.register(ConfigPage)  # type: ignore
def config(element, parent):
    return update(element, ConfigPage(parent, element['props']['config'], {'contents': []}))

@update.register(ConfigPage)  # type: ignore
def config(element, instance: ConfigPage):
    set_basic_props(instance, element['props'])
    return instance

@mount.register(TabbedConfigPage)  # type: ignore
def tabbedconfig(element, parent):
    return update(element, TabbedConfigPage(parent, element['props']['config'], {'contents': []}))

@update.register(TabbedConfigPage)  # type: ignore
def tabbedconfig(element, instance: TabbedConfigPage):
    set_basic_props(instance, element['props'])
    return instance

@mount.register(Console)  # type: ignore
def console(element, parent):
    return update(element, Console(parent, element['props']))

@update.register(Console)  # type: ignore
def console(element, instance: Console):
    set_basic_props(instance, element['props'])
    if 'show' in element['props']:
        instance.Show(element['props']['show'])
    return instance




from gooey.gui.components.widgets.textfield import TextField
from gooey.python_bindings import types as t



__ALL__ = ('CommandField',)

class CommandField(TextField):

    def getUiState(self) -> t.FormField:
        return t.Command(**super().getUiState())  # type: ignore



import os
import sys
import signal
import warnings
import textwrap
from gooey.python_bindings import argparse_to_json
from gooey.gui.util.quoting import quote
from gooey.python_bindings import constants
from gooey.python_bindings import gooey_decorator
from gooey.gui.util.functional import merge_dictionaries

default_layout = {
    'widgets': [{
      'type': 'CommandField',
      'required': True,
      'data': {
        'display_name': 'Enter Commands',
        'help': 'Enter command line arguments',
        'nargs': '',
        'commands': '',
        'choices': [],
        'default': None,
      }
    }],
}

# TODO: deprecate me
def create_from_parser(parser, source_path, **kwargs):

  run_cmd = kwargs.get('target')
  if run_cmd is None:
    if hasattr(sys, 'frozen'):
      run_cmd = quote(source_path)
    else:
      run_cmd = '{} -u {}'.format(quote(sys.executable), quote(source_path))

  build_spec = {**kwargs, 'target': run_cmd}

  if build_spec['monospace_display']:
      warnings.warn('Gooey Option `monospace_display` is a legacy option.\n'
                    'See the terminal_font_x options for more flexible control '
                    'over Gooey\'s text formatting')


  build_spec['program_description'] = build_spec['program_description'] or parser.description or ''

  layout_data = (argparse_to_json.convert(parser, **build_spec)
                   if build_spec['advanced']
                   else default_layout.items())

  build_spec.update(layout_data)

  if len(build_spec['widgets']) > 1:
    # there are subparsers involved
    build_spec['show_sidebar'] = True

  return build_spec






from argparse import ArgumentParser

from gooey import Events, Gooey


with open('tmp.txt', 'w') as f:
    import sys
    f.write(str(sys.argv))

def make_parser():
    parser = ArgumentParser()
    parser.add_argument('foo', type=int)
    return parser

@Gooey(use_events=[Events.VALIDATE_FORM])
def main():
    parser = make_parser()
    print(parser.parse_args())
    print('DONE')


if __name__ == '__main__':
    main()




from argparse import ArgumentParser
from gooey import Gooey, GooeyParser
import time
from time import sleep
import signal
import sys
import argparse



@Gooey()
def main():
    parser = GooeyParser(description='Process some data.')
    parser.add_argument('Choose Command Line',
                        choices=["--avhw --cbr 6000 --codec h265 --multipass none --preset quality --profile main10 --tier high --level 4.1 --output-depth 10 --vpp-edgelevel strength=9,threshold=12 --bframes 2 --ref 4 --gop-len 150 --lookahead 32 --aq --qp-init 1 --cuda-schedule auto --mv-precision q-pel --colorrange auto --colormatrix auto --colorprim auto --transfer auto --max-cll copy --master-display copy --videoformat auto --chromaloc auto --vpp-unsharp weight=1.5,radius=3 --vpp-gauss 3 --crop 0,0,0,0 --seek 0:00:00.000", "two"],
                        default="two",
                        widget='Dropdown'
                        )
    parser.parse_args()



if __name__ == '__main__':
    main()







import sys
import os

from time import sleep
from random import randint
from gooey import Gooey, GooeyParser


@Gooey(progress_regex=r"^progress: (?P<current>\d+)/(?P<total>\d+)$",
       progress_expr="current / total * 100",
       timing_options={
           'show_time_remaining': True,
           'hide_time_remaining_on_complete': False
       }
       )
def main():
    parser = GooeyParser(prog="example_progress_bar_3")
    parser.add_argument("steps", type=int, default=15)
    parser.add_argument("delay", type=int, default=1)
    parser.add_argument(
        "-v", "--something",
        metavar="* Something",
        help="Anything",
        default=1.0,
        widget="DecimalField")
    args = parser.parse_args(sys.argv[1:])

    for i in range(args.steps):
        print("progress: {}/{}".format(i + 1, args.steps))
        sys.stdout.flush()
        sleep(0.1)
    print("RESTING.....")
    sleep(1)
    for i in range(args.steps):
        print("progress: {}/{}".format(i + 1, args.steps))
        sys.stdout.flush()
        sleep(0.1)


if __name__ == "__main__":
    sys.exit(main())




import gooey
from gooey import Gooey

@Gooey
def main():
    parser = gooey.GooeyParser(description='Add two integers together, then subtract a decimal')
    parser.add_argument('--first', type=int, widget="IntegerField", gooey_options={'min': 5, 'max': 20})
    parser.add_argument('--second', type=int, widget="Slider", gooey_options={'min': 1, 'max': 10})
    parser.add_argument('--third', type=int, widget="DecimalField", gooey_options={'min': 0, 'max': 5.0})
    args = parser.parse_args()
    print(args.first + args.second - args.third)

main()


"""
Dynamic updates in Gooey

Simple example demonstrating feeding Gooey a dynamic list of file names at runtime
"""
import datetime
import json
import os
import sys
from argparse import ArgumentParser

from gooey import Gooey, GooeyParser
from gooey import options
from python_bindings.control import Gooey1

directory_error = (
    'Unable to create a folder for your save files!\n'
    'Make sure you\re running in a directory where you have write permissions!'
)

def letter_a(x):
    if x == 'a':
        return x
    else:
        raise Exception("Must be the letter 'a'")





@Gooey1(
    program_name="Using Dynamic Values",
    # poll_external_updates=True
)
def main():
    parser = ArgumentParser(description='An example of polling for updates at runtime')
    g = parser.add_argument_group()
    parser.add_argument(
        '--save',
        required=True,
        type=letter_a,
        metavar='Save Progress',
        # action='store_true',
        help='Take a snap shot of your current progress!'
    )

    kwargs = [
        {'action': 'store_const', 'type': int},
        {'action': 'store_true', 'type': int},
        {'action': 'store_false', 'type': int},
        {'action': 'append', 'type': int},
        {'action': 'append_const', 'type': int},
        {'action': 'count', 'type': int},
        {'action': 'count', 'type': int},
        {'action': 'count', 'type': int},
        {'action': 'count', 'type': int}
    ]
    # parser.add_argument('--a', action='store_const', type=int)
    # parser.add_argument('--b', action='', type=int)
    # parser.add_argument('--c', action='store_true', type=int)
    # parser.add_argument('--c', action='store_false', type=int)



    # parser.add_argument(
    #     '--load',
    #     metavar='Load Previous Save',
    #     help='Load a Previous save file',
    #     dest='filename',
    #     widget='Listbox',
    #     nargs='+',
    #     choices=list_savefiles(),
    #     gooey_options={
    #         'full_width': True,
    #         'validator': {
    #             'test': 'user_input != "Select Option"',
    #             'message': 'Choose a save file from the list'
    #         }
    #     }
    # )

    args = parser.parse_args()

    if args.save:
        save_file()
    else:
        for f in args.filename:
            read_file(os.path.join('saves', f))
            print('Finished reading file %s!' % f)


def read_file(path):
    with open(path, 'r') as f:
        print(f.read())


def list_savefiles():
    """ List all available files in the save directory """
    return list(sorted(os.listdir('saves'), reverse=True))


def show_error_modal(error_msg):
    """ Spawns a modal with error_msg"""
    # wx imported locally so as not to interfere with Gooey
    import wx
    app = wx.App()
    dlg = wx.MessageDialog(None, error_msg, 'Error', wx.ICON_ERROR)
    dlg.ShowModal()
    dlg.Destroy()


def save_file():
    """Save a plain ol' text file with some metadata in the title"""
    next_number = '{:04d}'.format(len(list_savefiles()) + 1)
    now = datetime.datetime.now().strftime('%b %d, %Y - %H.%M%p')
    filename = 'Save{}{}.save'.format(next_number, now)
    with open(os.path.join('saves', filename), 'w') as f:
        f.write("Hello World!")
    print('Saved {}!'.format(filename))


def mk_savedir():
    """
    Attempt to create a directory where we can store the user's save files
    """
    try:
        os.mkdir('saves')
    except IOError as e:
        if not e.winerror == 183:  # already exists
            show_error_modal(directory_error)
            sys.exit(1)


if __name__ == '__main__':
    mk_savedir()
    main()



"""
Because Gooey communicates with the host program
over stdin/out, we have to be able to differentiate what's
coming from gooey and structured, versus what is arbitrary
junk coming from the host's own logging.

To do this, we just prefix all written by gooey with the
literal string 'gooey::'. This lets us dig through all the
noisy stdout to find just the structured Gooey data we're
after.
"""

import json
from base64 import b64decode
from typing import Dict, Any

from gooey.python_bindings.schema import validate_public_state
from gooey.python_bindings.types import PublicGooeyState

prefix = 'gooey::'


def serialize_outbound(out: PublicGooeyState):
    """
    Attaches a prefix to whatever is about to be written
    to stdout so that we can differentiate it in the
    sea of other stdout writes
    """
    return prefix + json.dumps(out)


def deserialize_inbound(stdout: bytes, encoding):
    """
    Deserializes the incoming stdout payload after
    finding the relevant sections give the gooey prefix.
    e.g.
    std='foo\nbar\nstarting run\ngooey::{active_form: [...]}\n'
    => {active_form: [...]}
    """
    data = json.loads(stdout.decode(encoding).split(prefix)[-1])
    return validate_public_state(data)


def decode_payload(x):
    """
    To avoid quoting shenanigans, the json state sent from
    Gooey is b64ecoded for ease of CLI transfer. Argparse will
    usually barf when trying to parse json directly
    """
    return json.loads(b64decode(x))



"""
Utilities for patching Windows so that CTRL-C signals
can be received by process groups.

The best resource for understanding why this is required is the Python
Issue here: https://bugs.python.org/issue13368

**The official docs from both Python and Microsoft cannot be trusted due to inaccuracies**

The two sources directly conflict with regards to what signals are possible on Windows
under which circumstances.


Python's docs:
    https://bugs.python.org/issue13368
    On Windows, SIGTERM is an alias for terminate(). CTRL_C_EVENT and
    CTRL_BREAK_EVENT can be sent to processes started with a creationflags
    parameter which includes CREATE_NEW_PROCESS_GROUP.

Microsoft's docs:
    https://docs.microsoft.com/en-us/windows/console/generateconsolectrlevent
    Generates a CTRL+C signal. This signal cannot be generated for process groups.

Another piece of the puzzle:
    https://docs.microsoft.com/en-us/windows/console/handlerroutine#remarks
    Each console process has its own list of HandlerRoutine functions. Initially,
    this list contains only a default handler function that calls ExitProcess. A
    console process adds or removes additional handler functions by calling the
    SetConsoleCtrlHandler function, which does not affect the list of handler
    functions for other processes.

The most important line here is: "Initially, this list contains only the default
handler function that calls exit Process"

So, despite what Microsoft's docs for ctrlcevent state, it IS possible to send
the ctrl+c signal to process groups **IFF** the leader for that process group has
the appropriate handler installed.

We can solve this transparently in Gooey land by installing the handler on behalf
of the user when required.
"""
import sys
import signal
from textwrap import dedent


def requires_special_handler(platform, requested_signal):
    """
    Checks whether we need to attach additional handlers
    to this process to deal with ctrl-C signals
    """
    return platform.startswith("win") and requested_signal == signal.CTRL_C_EVENT


def install_handler():
    """
    Adds a Ctrl+C 'handler routine'. This allows the CTRL-C
    signal to be received even when the process was created in
    a new process group.

    See module docstring for additional info.
    """
    assert sys.platform.startswith("win")
    import ctypes

    help_msg = dedent('''
    Please open an issue here: https://github.com/chriskiehl/Gooey/issues
        
    GOOD NEWS: THERE IS A WORK AROUND: 
    
    Gooey only needs to attach special Windows handlers for the CTRL-C event. Consider using 
    the `CTRL_BREAK_EVENT` or `SIGTERM` instead to get past this error. 
    
    See the Graceful Shutdown docs for information on how to catch alternative signals. 
    
    https://github.com/chriskiehl/Gooey/tree/master/docs 
    ''')

    try:
        kernel32 = ctypes.WinDLL('kernel32')
        if kernel32.SetConsoleCtrlHandler(None, 0) == 0:
            raise Exception(dedent('''
            Gooey was unable to install the handler required for processing
            CTRL-C signals. This is a **very** unexpected failure. 
            ''') + help_msg)
    except OSError as e:
        raise OSError(dedent('''
            Gooey failed while trying to find the kernel32 module. 
            Gooey requires this module in order to attach handlers for CTRL-C signal. 
            Not being able to find this is **very** unexpected. 
            ''') + help_msg)



import json
from itertools import chain

from copy import deepcopy

from gooey.util.functional import compact
from typing import List, Optional

from gooey.gui.constants import VALUE_PLACEHOLDER
from gooey.gui.formatters import formatArgument
from gooey.python_bindings.types import FieldValue, Group, Item
from gooey.util.functional import merge  # type: ignore
from gooey.gui.state import FullGooeyState

'''
primary :: Target -> Command -> Array Arg -> Array Arg -> Boolean -> CliString
validateForm :: Target -> Command -> Array Arg -> Array Arg -> CliString
validateField :: Target -> Command -> Array Arg -> Array Arg -> ArgId -> CliString
completed :: Target -> Command -> FromState -> CliString
failed :: Target -> Command -> FromState -> CliString
fieldAction :: Target -> Command ->   

'''


def buildSuccessCmd(state: FullGooeyState):
    subcommand = state['subcommands'][state['activeSelection']]
    widgets = state['widgets'][subcommand]




def onSuccessCmd(target: str, subCommand: str, formState: List[str]) -> str:
    command = subCommand if not subCommand == '::gooey/default' else ''
    return f'{target} {command} --gooey-on-success {json.dumps(formState)}'


def onErrorCmd(target: str, subCommand: str, formState: List[str]) -> str:
    command = subCommand if not subCommand == '::gooey/default' else ''
    return f'{target} {command} --gooey-on-error {json.dumps(formState)}'


def formValidationCmd(target: str, subCommand: str, positionals: List[FieldValue], optionals: List[FieldValue]) -> str:
    positional_args = [cmdOrPlaceholderOrNone(x) for x in positionals]
    optional_args = [cmdOrPlaceholderOrNone(x) for x in optionals]
    command = subCommand if not subCommand == '::gooey/default' else ''
    return u' '.join(compact([
        target,
        command,
        *optional_args,
        '--gooey-validate-form',
        '--' if positional_args else '',
        *positional_args]))


def cliCmd(target: str,
           subCommand: str,
           positionals: List[FieldValue],
           optionals: List[FieldValue],
           suppress_gooey_flag=False) -> str:
    positional_args = [arg['cmd'] for arg in positionals]
    optional_args = [arg['cmd'] for arg in optionals]
    command = subCommand if not subCommand == '::gooey/default' else ''
    ignore_flag = '' if suppress_gooey_flag else '--ignore-gooey'
    return u' '.join(compact([
        target,
        command,
        *optional_args,
        ignore_flag,
        '--' if positional_args else '',
        *positional_args]))


def cmdOrPlaceholderOrNone(field: FieldValue) -> Optional[str]:
    # Argparse has a fail-fast-and-exit behavior for any missing
    # values. This poses a problem for dynamic validation, as we
    # want to collect _all_ errors to be more useful to the user.
    # As such, if there is no value currently available, we pass
    # through a stock placeholder values which allows GooeyParser
    # to handle it being missing without Argparse exploding due to
    # it actually being missing.
    if field['clitype'] == 'positional':
        return field['cmd'] or VALUE_PLACEHOLDER
    elif field['clitype'] != 'positional' and field['meta']['required']:
        # same rationale applies here. We supply the argument
        # along with a fixed placeholder (when relevant i.e. `store`
        # actions)
        return field['cmd'] or formatArgument(field['meta'], VALUE_PLACEHOLDER)
    else:
        # Optional values are, well, optional. So, like usual, we send
        # them if present or drop them if not.
        return field['cmd']


def buildCliString(target, subCommand, positional, optional, suppress_gooey_flag=False):
    positionals = deepcopy(positional)
    if positionals:
        positionals.insert(0, "--")

    arguments = ' '.join(compact(chain(optional, positionals)))

    if subCommand != '::gooey/default':
        arguments = u'{} {}'.format(subCommand, arguments)

    ignore_flag = '' if suppress_gooey_flag else '--ignore-gooey'
    return u'{} {} {}'.format(target, ignore_flag, arguments)



import wx  # type: ignore

from gooey.gui import formatters
from gooey.gui.components.widgets.bases import TextContainer
from gooey.gui.components.widgets.core.text_input import TextInput
from gooey.python_bindings import types as t

class TextField(TextContainer):
    widget_class = TextInput

    def getWidgetValue(self):
        return self.widget.getValue()

    def setValue(self, value):
        self.widget.setValue(str(value))

    def setPlaceholder(self, value):
        self.widget.SetHint(value)

    def formatOutput(self, metatdata, value):
        return formatters.general(metatdata, value)

    def syncUiState(self, state: t.TextField):  # type: ignore
        textctr: wx.TextCtrl = self.widget.widget
        textctr.SetValue(state['value'])
        textctr.SetHint(state['placeholder'])
        textctr.Enable(state['enabled'])
        self.Show(state['visible'])
        self.error.SetLabel(state['error'] or '')
        self.error.Show(state['error'] is not None and state['error'] is not '')
        self.Layout()



import rewx.components as c  # type: ignore
import wx  # type: ignore
import wx.html2  # type: ignore
from rewx import wsx, render  # type: ignore


def _html_window(html):
    return wsx(
        [c.Block, {'orient': wx.VERTICAL, 'flag': wx.EXPAND},
         [c.HtmlWindow, {'style': wx.TE_READONLY, 'flag': wx.EXPAND | wx.ALL,
                         'proportion': 1, 'value': html}]]
    )


class HtmlDialog(wx.Dialog):
    """
    A MessageDialog where the central contents are an HTML window
    customizable by the user.
    """
    def __init__(self, *args, **kwargs):
        caption = kwargs.pop('caption', '')
        html = kwargs.pop('html', '')
        super(HtmlDialog, self).__init__(None, *args, **kwargs)

        wx.InitAllImageHandlers()

        self.SetTitle(caption)
        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(render(_html_window(html), self), 1, wx.EXPAND)

        # in addition to creating the sizer, this actually attached
        # a few common handlers which makes it feel more dialog-y. Thus
        # it being done here rather than in rewx
        btnSizer = self.CreateStdDialogButtonSizer(wx.OK)
        sizer.Add(btnSizer, 0, wx.ALL | wx.EXPAND, 9)
        self.SetSizer(sizer)
        self.Layout()






import json
import unittest
from argparse import ArgumentParser
from contextlib import contextmanager
from pprint import pprint
from typing import Dict, List
from unittest.mock import MagicMock, patch

import sys
import shlex

from wx._core import CommandEvent

from gooey import GooeyParser
from python_bindings.coms import decode_payload, deserialize_inbound
from python_bindings.dynamics import patch_argument, check_value
from gooey.python_bindings import control
from gooey.python_bindings.parameters import gooey_params
from gooey.gui import state as s
from gooey.python_bindings.schema import validate_public_state
from python_bindings.types import FormField

from tests.harness import instrumentGooey

from gooey.tests import *


def custom_type(x):
    if x == '1234':
        return x
    else:
        raise Exception('KABOOM!')


class TestControl(unittest.TestCase):

    def tearDown(self):
        """
        Undoes the monkey patching after every tests
        """
        if hasattr(ArgumentParser, 'original_parse_args'):
            ArgumentParser.parse_args = ArgumentParser.original_parse_args

    def test_validate_form(self):
        """
        Testing the major validation cases we support.
        """
        writer = MagicMock()
        exit = MagicMock()
        monkey_patch = control.validate_form(gooey_params(), write=writer, exit=exit)
        ArgumentParser.original_parse_args = ArgumentParser.parse_args
        ArgumentParser.parse_args = monkey_patch

        parser = GooeyParser()
        # examples:
        # ERROR: mismatched builtin type
        parser.add_argument('a', type=int, gooey_options={'initial_value': 'not-an-int'})
        # ERROR: mismatched custom type
        parser.add_argument('b', type=custom_type, gooey_options={'initial_value': 'not-a-float'})
        # ERROR: missing required positional arg
        parser.add_argument('c')
        # ERROR: missing required 'optional' arg
        parser.add_argument('--oc', required=True)
        # VALID: This is one of the bizarre cases which are possible
        # but don't make much sense. It should pass through as valid
        # because there's no way for us to send a 'not present optional value'
        parser.add_argument('--bo', action='store_true', required=True)
        # ERROR: a required mutex group, with no args supplied.
        # Should flag all as missing.
        group = parser.add_mutually_exclusive_group(required=True)
        group.add_argument('--gp1-a', type=str)
        group.add_argument('--gp1-b', type=str)

        # ERROR: required mutex group with a default option but nothing
        # selected will still fail
        group2 = parser.add_mutually_exclusive_group(required=True)
        group2.add_argument('--gp2-a', type=str)
        group2.add_argument('--gp2-b', type=str, default='Heeeeyyyyy')

        # VALID: now, same as above, but now the option is actually enabled via
        # the initial selection. No error.
        group3 = parser.add_mutually_exclusive_group(required=True, gooey_options={'initial_selection': 1})
        group3.add_argument('--gp3-a', type=str)
        group3.add_argument('--gp3-b', type=str, default='Heeeeyyyyy')
        # VALID: optional mutex.
        group4 = parser.add_mutually_exclusive_group()
        group4.add_argument('--gp4-a', type=str)
        group4.add_argument('--gp4-b', type=str)
        # VALID: arg present and type satisfied
        parser.add_argument('ga', type=str, gooey_options={'initial_value': 'whatever'})
        # VALID: arg present and custom type satisfied
        parser.add_argument('gb', type=custom_type, gooey_options={'initial_value': '1234'})
        # VALID: optional
        parser.add_argument('--gc')

        # now we're adding the same
        with instrumentGooey(parser, target='test') as (app, frame, gapp):
            # we start off with no errors
            self.assertFalse(s.has_errors(gapp.fullState()))

            # now we feed our form-validation
            cmd = s.buildFormValidationCmd(gapp.fullState())
            asdf = shlex.split(cmd)[1:]
            parser.parse_args(shlex.split(cmd)[1:])
            assert writer.called
            assert exit.called


        result = deserialize_inbound(writer.call_args[0][0].encode('utf-8'), 'utf-8')
        # Host->Gooey communication is all done over the PublicGooeyState schema
        # as such, we coarsely validate it's shape here
        validate_public_state(result)

        # manually merging the two states back together
        nextState = s.mergeExternalState(gapp.fullState(), result)
        # and now we find that we have errors!
        self.assertTrue(s.has_errors(nextState))
        items = s.activeFormState(nextState)
        self.assertIn('invalid literal', get_by_id(items, 'a')['error'])
        self.assertIn('KABOOM!', get_by_id(items, 'b')['error'])
        self.assertIn('required', get_by_id(items, 'c')['error'])
        self.assertIn('required', get_by_id(items, 'oc')['error'])
        for item in get_by_id(items, 'group_gp1_a_gp1_b')['options']:
            self.assertIsNotNone(item['error'])
        for item in get_by_id(items, 'group_gp2_a_gp2_b')['options']:
            self.assertIsNotNone(item['error'])

        for item in get_by_id(items, 'group_gp3_a_gp3_b')['options']:
            self.assertIsNone(item['error'])
        # should be None, since this one was entirely optional
        for item in get_by_id(items, 'group_gp4_a_gp4_b')['options']:
            self.assertIsNone(item['error'])
        self.assertIsNone(get_by_id(items, 'bo')['error'])
        self.assertIsNone(get_by_id(items, 'ga')['error'])
        self.assertIsNone(get_by_id(items, 'gb')['error'])
        self.assertIsNone(get_by_id(items, 'gc')['error'])


    def test_subparsers(self):
        """
        Making sure that subparsers are handled correctly and
        all validations still work as expected.
        """
        writer = MagicMock()
        exit = MagicMock()
        monkey_patch = control.validate_form(gooey_params(), write=writer, exit=exit)
        ArgumentParser.original_parse_args = ArgumentParser.parse_args
        ArgumentParser.parse_args = monkey_patch

        def build_parser():
            # we build a new parser for each subtest
            # since we monkey patch the hell out of it
            # each time
            parser = GooeyParser()
            subs = parser.add_subparsers()
            foo = subs.add_parser('foo')
            foo.add_argument('a')
            foo.add_argument('b')
            foo.add_argument('p')

            bar = subs.add_parser('bar')
            bar.add_argument('a')
            bar.add_argument('b')
            bar.add_argument('z')
            return parser

        parser = build_parser()
        with instrumentGooey(parser, target='test') as (app, frame, gapp):
            with self.subTest('first subparser'):
                # we start off with no errors
                self.assertFalse(s.has_errors(gapp.fullState()))

                cmd = s.buildFormValidationCmd(gapp.fullState())
                parser.parse_args(shlex.split(cmd)[1:])
                assert writer.called
                assert exit.called

                result = deserialize_inbound(writer.call_args[0][0].encode('utf-8'), 'utf-8')
                nextState = s.mergeExternalState(gapp.fullState(), result)
                # by default, the subparser defined first, 'foo', is selected.
                self.assertIn('foo', nextState['forms'])
                # and we should find its attributes
                expected = {'a', 'b', 'p'}
                actual = {x['id'] for x in nextState['forms']['foo']}
                self.assertEqual(expected, actual)


        parser = build_parser()
        with instrumentGooey(parser, target='test') as (app, frame, gapp):
            with self.subTest('Second subparser'):
                # mocking a 'selection change' event to select
                # the second subparser
                event = MagicMock()
                event.Selection = 1
                gapp.handleSelectAction(event)

                # Flushing our events by running the main loop
                wx.CallLater(1, app.ExitMainLoop)
                app.MainLoop()

                cmd = s.buildFormValidationCmd(gapp.fullState())
                parser.parse_args(shlex.split(cmd)[1:])
                assert writer.called
                assert exit.called

                result = deserialize_inbound(writer.call_args[0][0].encode('utf-8'), 'utf-8')
                nextState = s.mergeExternalState(gapp.fullState(), result)
                # Now our second subparer, 'bar', should be present.
                self.assertIn('bar', nextState['forms'])
                # and we should find its attributes
                expected = {'a', 'b', 'z'}
                actual = {x['id'] for x in nextState['forms']['bar']}
                self.assertEqual(expected, actual)


    def test_ignore_gooey(self):
        parser = GooeyParser()
        subs = parser.add_subparsers()
        foo = subs.add_parser('foo')
        foo.add_argument('a')
        foo.add_argument('b')
        foo.add_argument('p')

        bar = subs.add_parser('bar')
        bar.add_argument('a')
        bar.add_argument('b')
        bar.add_argument('z')

        control.bypass_gooey(gooey_params())(parser)

def get_by_id(items: List[FormField], id: str):
    return [x for x in items if x['id'] == id][0]






import unittest
from argparse import ArgumentParser
from functools import wraps

from python_bindings.types import TimingOptions


# TODO:

# def decor(f=None, *gargs, **gkwargs):
#     @wraps(f)
#     def inner(*args, **kwargs):
#         print('hello from decorator', gargs, gkwargs)
#         # choose handler
#         # monkey patch parser
#
#         return f(*args, **kwargs)
#
#     def inner2(func):
#         return decor(func, *gargs, **gkwargs)
#
#     return inner if callable(f) else inner2
#
#
# def handle_success(params):
#     def parse_args(self: ArgumentParser, args=None, namespace=None):
#         return self._original_parse_args()
#     return parse_args
#
#
# # @decor
# def main(*args, **kwargs):
#     """Hellow world!!!!!"""
#     print('sup from main', args, kwargs)
#
#     # ArgumentParser._original_parse_args = ArgumentParser.parse_args
#     # ArgumentParser.parse_args = handle_success(ArgumentParser.parse_args)
#
#     parser = ArgumentParser()
#     parser.add_argument('-f', '--foo', help='is foo')
#     subs = parser.add_subparsers()
#     sp = subs.add_parser('hh')
#     sp.add_argument('-f', '--foo', help='sp.foo')
#     print(parser.parse_args(['hh', '-f', 'asdf']))
#
#     print(TimingOptions(show_time_remaining=True, hide_time_remaining_on_complete=True).hide_time_remaining_on_complete)
#
#
#
#
# class Testie(unittest.TestCase):
#
#     def test_thing(self, **kwargs):
#         print(main(1, 2))
#         print(help(main))


import math

from gooey import Gooey, GooeyParser

parser = GooeyParser()


@Gooey(program_name='Area of a Circle Program',
       menu=[{'name': 'Help', 'items': [{'type': 'Link', 'menuTitle': 'Visit Our Site',
                                         'url': 'https://github.com/chriskiehl/Gooey'}]}])
def main():
    parser.add_argument('-r', '--radius', type=float, help='Radius of a Circle')
    args = parser.parse_args()
    area = circle_area(args.radius)
    print(area)


def circle_area(r: float):
    return math.pi * r * r


if __name__ == '__main__':
    main()


"""
Python file for Processor test suite

Infinite loop which purposefully ignores
Keyboard Interrupts in order to continue
executing. The only way to kill it is via
SIGTERM family signals.
"""

import time
import sys

if sys.platform.startswith('win'):
    import ctypes
    kernel32 = ctypes.WinDLL('kernel32')
    kernel32.SetConsoleCtrlHandler(None, 0)


while True:
    try:
        print(time.time())
        time.sleep(0.1)
    except KeyboardInterrupt:
        # Ignored!
        print("INTERRUPT")



from argparse import _StoreAction

from gooey import Gooey, GooeyParser
from gooey import Gooey
from gooey import Events


def foo(*args):
    return [1,2,3]


def mything(x):
    num = int(x)
    if num > 0 and num < 10:
        return num
    else:
        raise ValueError('Must be between 0 and 10 inclusive')


class MyAction(_StoreAction):
    def __call__(self, *args, **kwargs):
        # print('Hello!!!')
        super(MyAction, self).__call__(*args, **kwargs)


@Gooey(program_name="Gooey!", dump_build_config=True,
        # cli='user-code -z "False" --ignore-gooey --gooey-validate-form --'.split()
        # cli='"C:\\Users\\Chris\\Documents\\Gooey\\venv368\\Scripts\\python.exe" -u "C:/Users/Chris/Documents/Gooey/gooey/examples/a1.py" user-code -p "1234" --gooey-validate-form',
       use_events=[Events.VALIDATE_FORM]
        )
def main():
    parser = GooeyParser(prog="Gooey",
                         description="Gooey turns your CLI apps into beautiful GUIs!",
                         )
    parser.add_argument('-f', '--foo', help='root', gooey_options={
        'root': 'root f!!'
    })
    subs = parser.add_subparsers(title='gooey-managed')
    clientvvv = subs.add_parser('user-smode')
    clientp = subs.add_parser('user-code')
    clientp.add_argument('-f', '--foo',
                         type=mything,
                         action=MyAction,
                         help='subbie',
                         widget='DirChooser',
                         gooey_options={
        'sub': 'sub f!!'
    })
    clientp.add_argument('-z', '--zzz')
    clientp.add_argument('-d', '--dope', type=int, widget='DirChooser')
    clientp.add_argument('-u', '--uope', type=mything, gooey_options={
        'validator': {
            'test': 'int(user_input) < 2',
            'message': 'some helpful message'
    }
    })
    clientp.add_argument('-p', type=int, choices=foo(),  gooey_options={
    })

    # args = parser.parse_args(['user-code', '-f', 'yo yo yo', '-d', '123'])
    # args = parser.parse_args('user-code -p 1234 --gooey-validate-form'.split())
    args = parser.parse_args()
    import sys
    a = 10
    print(args)
    # print('DUN got me fucked up!')
    # print('DUN GOOFED')
    # import sys
    # sys.exit(-80085)


if __name__ == '__main__':
    conf = main()
    # print("Done")


import sys
from json import JSONDecodeError

import six
import wx  # type: ignore

from gooey import Events
from gooey.gui import events
from gooey.gui import host
from gooey.gui import state as s
from gooey.gui.application.components import RHeader, ProgressSpinner, ErrorWarning, RTabbedLayout, \
    RSidebar, RFooter
from gooey.gui.components import modals
from gooey.gui.components.config import ConfigPage
from gooey.gui.components.config import TabbedConfigPage
from gooey.gui.components.console import Console
from gooey.gui.components.menubar import MenuBar
from gooey.gui.lang.i18n import _
from gooey.gui.processor import ProcessController
from gooey.gui.pubsub import pub
from gooey.gui.state import FullGooeyState
from gooey.gui.state import initial_state, ProgressEvent, TimingEvent
from gooey.gui.util.wx_util import transactUI, callafter
from gooey.python_bindings import constants
from gooey.python_bindings.dynamics import unexpected_exit_explanations, \
    deserialize_failure_explanations
from gooey.python_bindings.types import PublicGooeyState
from gooey.python_bindings.types import Try
from gooey.util.functional import assoc
from gooey.gui.util.time import Timing
from rewx import components as c  # type: ignore
from rewx import wsx  # type: ignore
from rewx.core import Component, Ref  # type: ignore


class RGooey(Component):
    """
    Main Application container for Gooey.

    State Management
    ----------------

    Pending further refactor, state is tracked in two places:
    1. On this instance (React style)
    2. In the WX Form Elements themselves[0]

    As needed, these two states are merged to form the `FullGooeyState`, which
    is the canonical state object against which all logic runs.


    Dynamic Updates
    ---------------




    [0] this is legacy and will (eventually) be refactored away

    """
    def __init__(self, props):
        super().__init__(props)
        self.frameRef = Ref()
        self.consoleRef = Ref()
        self.configRef = Ref()

        self.buildSpec = props
        self.state = initial_state(props)
        self.headerprops = lambda state: {
            'background_color': self.buildSpec['header_bg_color'],
            'title': state['title'],
            'show_title': state['header_show_title'],
            'subtitle': state['subtitle'],
            'show_subtitle': state['header_show_subtitle'],
            'flag': wx.EXPAND,
            'height': self.buildSpec['header_height'],
            'image_uri': state['image'],
            'image_size': (six.MAXSIZE, self.buildSpec['header_height'] - 10)}

        self.fprops = lambda state: {
            'buttons': state['buttons'],
            'progress': state['progress'],
            'timing': state['timing'],
            'bg_color': self.buildSpec['footer_bg_color'],
            'flag': wx.EXPAND,
        }
        self.clientRunner = ProcessController.of(self.buildSpec)
        self.timer = None


    def component_did_mount(self):
        pub.subscribe(events.WINDOW_START, self.onStart)
        pub.subscribe(events.WINDOW_RESTART, self.onStart)
        pub.subscribe(events.WINDOW_STOP, self.handleInterrupt)
        pub.subscribe(events.WINDOW_CLOSE, self.handleClose)
        pub.subscribe(events.WINDOW_CANCEL, self.handleCancel)
        pub.subscribe(events.WINDOW_EDIT, self.handleEdit)
        pub.subscribe(events.CONSOLE_UPDATE, self.consoleRef.instance.logOutput)
        pub.subscribe(events.EXECUTION_COMPLETE, self.handleComplete)
        pub.subscribe(events.PROGRESS_UPDATE, self.updateProgressBar)
        pub.subscribe(events.TIME_UPDATE, self.updateTime)
        # # Top level wx close event
        frame: wx.Frame = self.frameRef.instance
        frame.Bind(wx.EVT_CLOSE, self.handleClose)
        frame.SetMenuBar(MenuBar(self.buildSpec))
        self.timer = Timing(frame)

        if self.state['fullscreen']:
            frame.ShowFullScreen(True)

        if self.state['show_preview_warning'] and not 'unittest' in sys.modules.keys():
            wx.MessageDialog(None, caption='YOU CAN DISABLE THIS MESSAGE',
                             message="""
                This is a preview build of 1.2.0! There may be instability or 
                broken functionality. If you encounter any issues, please open an issue 
                here: https://github.com/chriskiehl/Gooey/issues 
                
                The current stable version is 1.0.8. 
                
                NOTE! You can disable this message by setting `show_preview_warning` to False. 
                
                e.g. 
                `@Gooey(show_preview_warning=False)`
                """).ShowModal()

    def getActiveConfig(self):
        return [item
                for child in self.configRef.instance.Children
                # we descend down another level of children  to account
                # for Notebook layouts (which have wrapper objects)
                for item in [child] + list(child.Children)
                if isinstance(item, ConfigPage)
                or isinstance(item, TabbedConfigPage)][self.state['activeSelection']]

    def getActiveFormState(self):
        """
        This boiler-plate and manual interrogation of the UIs
        state is required until we finish porting the Config Form
        over to rewx (which is a battle left for another day given
        its complexity)
        """
        return self.getActiveConfig().getFormState()


    def fullState(self):
        """
        Re: final porting is a to do. For now we merge the UI
        state into the main tracked state.
        """
        formState = self.getActiveFormState()
        return s.combine(self.state, self.props, formState)


    def onStart(self, *args, **kwargs):
        """
        Dispatches the start behavior.
        """
        if Events.VALIDATE_FORM in self.state['use_events']:
            self.runAsyncValidation()
        else:
            self.startRun()


    def startRun(self):
        """
        Kicks off a run by invoking the host's code
        and pumping its stdout to Gooey's Console window.
        """
        state = self.fullState()
        if state['clear_before_run']:
            self.consoleRef.instance.Clear()
        self.set_state(s.consoleScreen(_, state))
        self.clientRunner.run(s.buildInvocationCmd(state))
        self.timer.start()
        self.frameRef.instance.Layout()
        for child in self.frameRef.instance.Children:
            child.Layout()


    def syncExternalState(self, state: FullGooeyState):
        """
        Sync the UI's state to what the host program has requested.
        """
        self.getActiveConfig().syncFormState(s.activeFormState(state))
        self.frameRef.instance.Layout()
        for child in self.frameRef.instance.Children:
            child.Layout()


    def handleInterrupt(self, *args, **kwargs):
        if self.shouldStopExecution():
            self.clientRunner.stop()

    def handleComplete(self, *args, **kwargs):
        self.timer.stop()
        if self.clientRunner.was_success():
            self.handleSuccessfulRun()
            if Events.ON_SUCCESS in self.state['use_events']:
                self.runAsyncExternalOnCompleteHandler(was_success=True)
        else:
            self.handleErrantRun()
            if Events.ON_ERROR in self.state['use_events']:
                self.runAsyncExternalOnCompleteHandler(was_success=False)

    def handleSuccessfulRun(self):
        if self.state['return_to_config']:
            self.set_state(s.editScreen(_, self.state))
        else:
            self.set_state(s.successScreen(_, self.state))
            if self.state['show_success_modal']:
                wx.CallAfter(modals.showSuccess)


    def handleErrantRun(self):
        if self.clientRunner.wasForcefullyStopped:
            self.set_state(s.interruptedScreen(_, self.state))
        else:
            self.set_state(s.errorScreen(_, self.state))
            if self.state['show_failure_modal']:
                wx.CallAfter(modals.showFailure)


    def successScreen(self):
        strings = {'title': _('finished_title'), 'subtitle': _('finished_msg')}
        self.set_state(s.success(self.state, strings, self.buildSpec))


    def handleEdit(self, *args, **kwargs):
        self.set_state(s.editScreen(_, self.state))

    def handleCancel(self, *args, **kwargs):
        if modals.confirmExit():
            self.handleClose()

    def handleClose(self, *args, **kwargs):
        """Stop any actively running client program, cleanup the top
        level WxFrame and shutdown the current process"""
        # issue #592 - we need to run the same onStopExecution machinery
        # when the exit button is clicked to ensure everything is cleaned
        # up correctly.
        frame: wx.Frame = self.frameRef.instance
        if self.clientRunner.running():
            if self.shouldStopExecution():
                self.clientRunner.stop()
                frame.Destroy()
                # TODO: NOT exiting here would allow
                # spawing the gooey to input params then
                # returning control to the CLI
                sys.exit()
        else:
            frame.Destroy()
            sys.exit()

    def shouldStopExecution(self):
        return not self.state['show_stop_warning'] or modals.confirmForceStop()

    def updateProgressBar(self, *args, progress=None):
        self.set_state(s.updateProgress(self.state, ProgressEvent(progress=progress)))

    def updateTime(self, *args, elapsed_time=None, estimatedRemaining=None, **kwargs):
        event = TimingEvent(elapsed_time=elapsed_time, estimatedRemaining=estimatedRemaining)
        self.set_state(s.updateTime(self.state, event))

    def handleSelectAction(self, event):
        self.set_state(assoc(self.state, 'activeSelection', event.Selection))


    def runAsyncValidation(self):
        def handleHostResponse(hostState: PublicGooeyState):
            self.set_state(s.finishUpdate(self.state))
            currentState = self.fullState()
            self.syncExternalState(s.mergeExternalState(currentState, hostState))
            if not s.has_errors(self.fullState()):
                self.startRun()
            else:
                self.set_state(s.editScreen(_, s.show_alert(self.fullState())))

        def onComplete(result: Try[PublicGooeyState]):
            result.onSuccess(handleHostResponse)
            result.onError(self.handleHostError)

        self.set_state(s.beginUpdate(self.state))
        fullState = self.fullState()
        host.communicateFormValidation(fullState, callafter(onComplete))


    def runAsyncExternalOnCompleteHandler(self, was_success):
        def handleHostResponse(hostState):
            if hostState:
                self.syncExternalState(s.mergeExternalState(self.fullState(), hostState))

        def onComplete(result: Try[PublicGooeyState]):
            result.onError(self.handleHostError)
            result.onSuccess(handleHostResponse)

        if was_success:
            host.communicateSuccessState(self.fullState(), callafter(onComplete))
        else:
            host.communicateErrorState(self.fullState(), callafter(onComplete))


    def handleHostError(self, ex):
        """
        All async errors get pumped here where we dump out the
        error and they hopefully provide a lot of helpful debugging info
        for the user.
        """
        try:
            self.set_state(s.errorScreen(_, self.state))
            self.consoleRef.instance.appendText(str(ex))
            self.consoleRef.instance.appendText(str(getattr(ex, 'output', '')))
            self.consoleRef.instance.appendText(str(getattr(ex, 'stderr', '')))
            raise ex
        except JSONDecodeError as e:
            self.consoleRef.instance.appendText(deserialize_failure_explanations)
        except Exception as e:
            self.consoleRef.instance.appendText(unexpected_exit_explanations)
        finally:
            self.set_state({**self.state, 'fetchingUpdate': False})


    def render(self):
        return wsx(
            [c.Frame, {'title': self.buildSpec['program_name'],
                       'background_color': self.buildSpec['body_bg_color'],
                       'double_buffered': True,
                       'min_size': (400, 300),
                       'icon_uri': self.state['images']['programIcon'],
                       'size': self.buildSpec['default_size'],
                       'ref': self.frameRef},
             [c.Block, {'orient': wx.VERTICAL},
              [RHeader, self.headerprops(self.state)],
              [c.StaticLine, {'style': wx.LI_HORIZONTAL, 'flag': wx.EXPAND}],
              [ProgressSpinner, {'show': self.state['fetchingUpdate']}],
              [ErrorWarning, {'show': self.state['show_error_alert'],
                              'uri': self.state['images']['errorIcon']}],
              [Console, {**self.buildSpec,
                         'flag': wx.EXPAND,
                         'proportion': 1,
                         'show': self.state['screen'] == 'CONSOLE',
                         'ref': self.consoleRef}],
              [RTabbedLayout if self.buildSpec['navigation'] == constants.TABBED else RSidebar,
               {'bg_color': self.buildSpec['sidebar_bg_color'],
                'label': 'Some Action!',
                'tabbed_groups': self.buildSpec['tabbed_groups'],
                'show_sidebar': self.state['show_sidebar'],
                'ref': self.configRef,
                'show': self.state['screen'] == 'FORM',
                'activeSelection': self.state['activeSelection'],
                'options': list(self.buildSpec['widgets'].keys()),
                'on_change': self.handleSelectAction,
                'config': self.buildSpec['widgets'],
                'flag': wx.EXPAND,
                'proportion': 1}],
              [c.StaticLine, {'style': wx.LI_HORIZONTAL, 'flag': wx.EXPAND}],
              [RFooter, self.fprops(self.state)]]]
        )







from argparse import ArgumentParser

from gooey import Events, Gooey, GooeyParser
from gooey import types as t


with open('tmp.txt', 'w') as f:
    import sys
    f.write(str(sys.argv))



def handle_success(args, state: t.PublicGooeyState):
    field = state['active_form'][0]
    field['value'] = 'success'
    return {**state, 'active_form': [field]}


def handle_error(args, state: t.PublicGooeyState):
    field = state['active_form'][0]
    field['value'] = 'error'
    return {**state, 'active_form': [field]}


def make_parser():
    parser = GooeyParser(on_error=handle_error, on_success=handle_success)
    parser.add_argument('foo')
    return parser

@Gooey(use_events=[Events.ON_ERROR, Events.ON_SUCCESS])
def main():
    parser = make_parser()
    args = parser.parse_args()
    if args.foo == 'fail':
        raise Exception('EXCEPTION')
    print('DONE')


if __name__ == '__main__':
    main()




from gooey import Gooey, Events
from argparse import ArgumentParser

def must_be_exactly_ten(value):
    number = int(value)
    if number == 10:
        return number
    else:
        raise TypeError("Hey! you need to provide exactly the number 10!")

@Gooey(program_name='Validation Example', use_events=[Events.VALIDATE_FORM])
def main():
    parser = ArgumentParser(description="Checkout this validation!")
    parser.add_argument('--ten', metavar='This field should be 10', type=must_be_exactly_ten)
    args = parser.parse_args()
    print(args)

if __name__ == '__main__':
    main()





import json
from base64 import b64encode
from typing import Optional, List, Dict, Any, Union, Callable

from typing_extensions import TypedDict
import wx

from gooey.gui import events
from gooey.gui.lang.i18n import _
from gooey.python_bindings.types import GooeyParams, Item, Group, TopLevelParser, EnrichedItem, \
    FieldValue
from gooey.util.functional import associn, assoc, associnMany, compact
from gooey.gui.formatters import formatArgument
from gooey.python_bindings.types import FormField
from gooey.gui.constants import VALUE_PLACEHOLDER
from gooey.gui.formatters import add_placeholder
from gooey.python_bindings.types import CommandPieces, PublicGooeyState


class TimingEvent(TypedDict):
    elapsed_time: Optional[str]
    estimatedRemaining: Optional[str]

class ProgressEvent(TypedDict):
    progress: Optional[int]

class ButtonState(TypedDict):
    id: str
    style: str
    label_id: str
    show: bool
    enabled: bool

class ProgressState(TypedDict):
    show: bool
    range: int
    value: int

class TimingState(TypedDict):
    show: bool
    elapsedTime: Optional[str]
    estimated_remaining: Optional[str]

class GooeyState(GooeyParams):
    fetchingUpdate: bool
    screen: str
    title: str
    subtitle: str
    images: Dict[str, str]
    image: str
    buttons: List[ButtonState]
    progress: ProgressState
    timing: TimingState
    subcommands: List[str]
    activeSelection: int
    show_error_alert: bool

class FullGooeyState(GooeyState):
    forms: Dict[str, List[FormField]]
    widgets: Dict[str, Dict[str, Any]]





def extract_items(groups: List[Group]) -> List[Item]:
    if not groups:
        return []
    group = groups[0]
    return group['items'] \
           + extract_items(groups[1:]) \
           + extract_items(group['groups'])


def widgets(descriptor: TopLevelParser) -> List[Item]:
    return extract_items(descriptor['contents'])


def enrichValue(formState: List[FormField], items: List[Item]) -> List[EnrichedItem]:
    formIndex = {k['id']:k for k in formState}
    return [EnrichedItem(field=formIndex[item['id']], **item) for item in items]  # type: ignore


def positional(items: List[Union[Item, EnrichedItem]]):
    return [item for item in items if item['cli_type'] == 'positional']


def optional(items: List[Union[Item, EnrichedItem]]):
    return [item for item in items if item['cli_type'] != 'positional']


def cli_pieces(state: FullGooeyState, formatter=formatArgument) -> CommandPieces:
    parserName = state['subcommands'][state['activeSelection']]
    parserSpec = state['widgets'][parserName]
    formState = state['forms'][parserName]
    subcommand = parserSpec['command'] if parserSpec['command'] != '::gooey/default' else ''
    items = enrichValue(formState, widgets(parserSpec))
    positional_args = [formatter(item) for item in positional(items)]  # type: ignore
    optional_args = [formatter(item) for item in optional(items)]      # type: ignore
    ignoreFlag = '' if state['suppress_gooey_flag'] else '--ignore-gooey'
    return CommandPieces(
        target=state['target'],
        subcommand=subcommand,
        positionals=compact(positional_args),
        optionals=compact(optional_args),
        ignoreFlag=ignoreFlag
    )


def activeFormState(state: FullGooeyState):
    subcommand = state['subcommands'][state['activeSelection']]
    return state['forms'][subcommand]


def buildInvocationCmd(state: FullGooeyState):
    pieces = cli_pieces(state)
    return u' '.join(compact([
        pieces.target,
        pieces.subcommand,
        *pieces.optionals,
        pieces.ignoreFlag,
        '--' if pieces.positionals else '',
        *pieces.positionals]))


def buildFormValidationCmd(state: FullGooeyState):
    pieces = cli_pieces(state, formatter=cmdOrPlaceholderOrNone)
    serializedForm = json.dumps({'active_form': activeFormState(state)})
    b64ecoded = b64encode(serializedForm.encode('utf-8'))
    return ' '.join(compact([
        pieces.target,
        pieces.subcommand,
        *pieces.optionals,
        '--gooey-validate-form',
        '--gooey-state ' + b64ecoded.decode('utf-8'),
        '--' if pieces.positionals else '',
        *pieces.positionals]))


def buildOnCompleteCmd(state: FullGooeyState, was_success: bool):
    pieces = cli_pieces(state)
    serializedForm = json.dumps({'active_form': activeFormState(state)})
    b64ecoded = b64encode(serializedForm.encode('utf-8'))
    return u' '.join(compact([
        pieces.target,
        pieces.subcommand,
        *pieces.optionals,
        '--gooey-state ' + b64ecoded.decode('utf-8'),
        '--gooey-run-is-success' if was_success else '--gooey-run-is-failure',
        '--' if pieces.positionals else '',
        *pieces.positionals]))


def buildOnSuccessCmd(state: FullGooeyState):
    return buildOnCompleteCmd(state, True)

def buildOnErrorCmd(state: FullGooeyState):
    return buildOnCompleteCmd(state, False)


def cmdOrPlaceholderOrNone(item: EnrichedItem) -> Optional[str]:
    # Argparse has a fail-fast-and-exit behavior for any missing
    # values. This poses a problem for dynamic validation, as we
    # want to collect _all_ errors to be more useful to the user.
    # As such, if there is no value currently available, we pass
    # through a stock placeholder values which allows GooeyParser
    # to handle it being missing without Argparse exploding due to
    # it actually being missing.
    if item['cli_type'] == 'positional':
        return formatArgument(item) or VALUE_PLACEHOLDER
    elif item['cli_type'] != 'positional' and item['required']:
        # same rationale applies here. We supply the argument
        # along with a fixed placeholder (when relevant i.e. `store`
        # actions)
        return formatArgument(item) or formatArgument(assoc(item, 'field', add_placeholder(item['field'])))
    else:
        # Optional values are, well, optional. So, like usual, we send
        # them if present or drop them if not.
        return formatArgument(item)




def combine(state: GooeyState, params: GooeyParams, formState: List[FormField]) -> FullGooeyState:
    """
    I'm leaving the refactor of the form elements to another day.
    For now, we'll just merge in the state of the form fields as tracked
    in the UI into the main state blob as needed.
    """
    subcommand = list(params['widgets'].keys())[state['activeSelection']]
    return FullGooeyState(**{
        **state,
        **params,
        'forms': {subcommand: formState}
    })


def enable_buttons(state, to_enable: List[str]):
    updated = [{**btn, 'enabled': btn['label_id'] in to_enable}
               for btn in state['buttons']]
    return assoc(state, 'buttons', updated)



def activeCommand(state, params: GooeyParams):
    """
    Retrieve the active sub-parser command as determined by the
    current selection.
    """
    return list(params['widgets'].keys())[state['activeSelection']]


def mergeExternalState(state: FullGooeyState, extern: PublicGooeyState) -> FullGooeyState:
    # TODO: insane amounts of helpful validation
    subcommand = state['subcommands'][state['activeSelection']]
    formItems: List[FormField] = state['forms'][subcommand]
    hostForm: List[FormField] = extern['active_form']
    return associn(state, ['forms', subcommand], hostForm)


def show_alert(state: FullGooeyState):
    return assoc(state, 'show_error_alert', True)

def has_errors(state: FullGooeyState):
    """
    Searches through the form elements (including down into
    RadioGroup's internal options to find the presence of
    any errors.
    """
    return any([item['error'] or any(x['error'] for x in item.get('options', []))
                for items in state['forms'].values()
                for item in items])


def initial_state(params: GooeyParams) -> GooeyState:
    buttons = [
        ('cancel', events.WINDOW_CANCEL, wx.ID_CANCEL),
        ('start', events.WINDOW_START, wx.ID_OK),
        ('stop', events.WINDOW_STOP, wx.ID_OK),
        ('edit', events.WINDOW_EDIT,wx.ID_OK),
        ('restart', events.WINDOW_RESTART, wx.ID_OK),
        ('close', events.WINDOW_CLOSE, wx.ID_OK),
    ]
    # helping out the type system
    params: Dict[str, Any] = params
    return GooeyState(
        **params,
        fetchingUpdate=False,
        screen='FORM',
        title=params['program_name'],
        subtitle=params['program_description'],
        image=params['images']['configIcon'],
        buttons=[ButtonState(
            id=event_id,
            style=style,
            label_id=label,
            show=label in ('cancel', 'start'),
            enabled=True)
            for label, event_id, style in buttons],
        progress=ProgressState(
            show=False,
            range=100,
            value=0 if params['progress_regex'] else -1
        ),
        timing=TimingState(
            show=False,
            elapsed_time=None,
            estimatedRemaining=None,
        ),
        show_error_alert=False,
        subcommands=list(params['widgets'].keys()),
        activeSelection=0
    )

def header_props(state, params):
    return {
            'background_color': params['header_bg_color'],
            'title': params['program_name'],
            'subtitle': params['program_description'],
            'height': params['header_height'],
            'image_uri': ims['images']['configIcon'],
            'image_size': (six.MAXSIZE, params['header_height'] - 10)
    }


def form_page(state):
    return {
        **state,
        'buttons': [{**btn, 'show': btn['label_id'] in ('start', 'cancel')}
                    for btn in state['buttons']]
    }


def consoleScreen(_: Callable[[str], str], state: GooeyState):
    return {
        **state,
        'screen': 'CONSOLE',
        'title': _("running_title"),
        'subtitle': _('running_msg'),
        'image': state['images']['runningIcon'],
        'buttons': [{**btn,
                     'show': btn['label_id'] == 'stop',
                     'enabled': True}
                    for btn in state['buttons']],
        'progress': {
            'show': not state['disable_progress_bar_animation'],
            'range': 100,
            'value': 0 if state['progress_regex'] else -1
        },
        'timing': {
            'show': state['timing_options']['show_time_remaining'],
            'elapsed_time': None,
            'estimatedRemaining': None
        },
        'show_error_alert': False
     }


def editScreen(_: Callable[[str], str], state: FullGooeyState):
    use_buttons = ('cancel', 'start')
    return associnMany(
        state,
        ('screen', 'FORM'),
        ('buttons', [{**btn,
                      'show': btn['label_id'] in use_buttons,
                      'enabled': True}
                     for btn in state['buttons']]),
        ('image', state['images']['configIcon']),
        ('title', state['program_name']),
        ('subtitle', state['program_description']))


def beginUpdate(state: GooeyState):
    return {
        **enable_buttons(state, ['cancel']),
        'fetchingUpdate': True
    }

def finishUpdate(state: GooeyState):
    return {
        **enable_buttons(state, ['cancel', 'start']),
        'fetchingUpdate': False
    }


def finalScreen(_: Callable[[str], str], state: GooeyState) -> GooeyState:
    use_buttons = ('edit', 'restart', 'close')
    return associnMany(
        state,
        ('screen', 'CONSOLE'),
        ('buttons', [{**btn,
                      'show': btn['label_id'] in use_buttons,
                      'enabled': True}
                     for btn in state['buttons']]),
        ('image', state['images']['successIcon']),
        ('title', _('finished_title')),
        ('subtitle', _('finished_msg')),
        ('progress.show', False),
        ('timing.show', not state['timing_options']['hide_time_remaining_on_complete']))


def successScreen(_: Callable[[str], str], state: GooeyState) -> GooeyState:
    return associnMany(
        finalScreen(_, state),
        ('image', state['images']['successIcon']),
        ('title', _('finished_title')),
        ('subtitle', _('finished_msg')))


def errorScreen(_: Callable[[str], str], state: GooeyState) -> GooeyState:
    return associnMany(
        finalScreen(_, state),
        ('image', state['images']['errorIcon']),
        ('title', _('finished_title')),
        ('subtitle', _('finished_error')))


def interruptedScreen(_: Callable[[str], str], state: GooeyState):
    next_state = errorScreen(_, state) if state['force_stop_is_error'] else successScreen(_, state)
    return assoc(next_state, 'subtitle', _('finished_forced_quit'))


def updateProgress(state, event: ProgressEvent):
    return associn(state, ['progress', 'value'], event['progress'] or 0)


def updateTime(state, event):
    return associnMany(
        state,
        ('timing.elapsed_time', event['elapsed_time']),
        ('timing.estimatedRemaining', event['estimatedRemaining'])
    )






def update_time(state, event: TimingEvent):
    return {
        **state,
        'timer': {
            **state['timer'],
            'elapsed_time': event['elapsed_time'],
            'estimatedRemaining': event['estimatedRemaining']
        }
    }


def present_time(timer):
    estimate_time_remaining = timer['estimatedRemaining']
    elapsed_time_value = timer['elapsed_time']
    if elapsed_time_value is None:
        return ''
    elif estimate_time_remaining is not None:
        return f'{elapsed_time_value}<{estimate_time_remaining}'
    else:
        return f'{elapsed_time_value}'
